from pytest import fixture

from py_rpautom.desktop_utils import (
    _aplicacao,
    capturar_texto,
    clicar,
    coletar_dado_selecionado,
    coletar_dados_selecao,
    coletar_situacao_janela,
    digitar,
    encerrar_app,
    esta_visivel,
    fechar_janela,
    iniciar_app,
    _localizar_elemento,
    maximizar_janela,
    minimizar_janela,
    restaurar_janela,
    selecionar_em_campo_selecao,
    selecionar_menu,
)
from py_rpautom.python_utils import (
    caminho_existente,
    criar_arquivo_texto,
    criar_pasta,
    excluir_arquivo,
    excluir_pasta,
    caminho_existente,
)


@fixture
def aplicacao_test():
    return _aplicacao()


@fixture
def executavel_mouseclicker():
    executavel_path = 'tests/mouseclicker.exe'
    return executavel_path


@fixture
def executavel_notepad():
    executavel_path = 'notepad'
    return executavel_path


@fixture
def caminho_campo():
    caminho_campo = 'Free Mouse Clicker->Start'
    return caminho_campo


@fixture
def valor():
    valor = 5
    return valor


@fixture
def iniciar_app_test(executavel_mouseclicker):
    return iniciar_app(executavel_mouseclicker)


@fixture
def encerrar_app_test(executavel_mouseclicker):
    encerrar_app(executavel_mouseclicker)
    return True


@fixture
def contexto_mouseclicker(executavel_mouseclicker):
    app = iniciar_app(executavel_mouseclicker)
    yield app
    encerrar_app(executavel_mouseclicker)
    return app


@fixture
def contexto_notepad(executavel_notepad):
    app = iniciar_app(executavel_notepad)
    yield app
    encerrar_app(executavel_notepad)
    return app


@fixture
def digitar_test(caminho_campo, valor):
    return digitar(caminho_campo, valor)


@fixture
def _localizar_elemento_estatico_test(caminho_campo):
    return _localizar_elemento(caminho_campo, estatico=True)


@fixture
def _localizar_elemento_dinamico_test(caminho_campo):
    return _localizar_elemento(caminho_campo, estatico=False)


@fixture
def capturar_texto_test(caminho_campo):
    capturar_texto(caminho_campo)


@fixture
def clicar_test(caminho_campo):
    return clicar(caminho_campo)


@fixture
def esta_visivel_test():
    nome_janela = 'Free Mouse Clicker'
    return esta_visivel(nome_janela)


@fixture
def coletar_situacao_janela_normal_test():
    nome_janela = 'Free Mouse Clicker'
    return coletar_situacao_janela(nome_janela)


@fixture
def coletar_situacao_janela_minimizada_test():
    nome_janela = 'Free Mouse Clicker'
    minimizar_janela(nome_janela)
    return coletar_situacao_janela(nome_janela)


@fixture
def coletar_situacao_janela_maximizada_test():
    nome_janela = 'Free Mouse Clicker'
    maximizar_janela(nome_janela)
    return coletar_situacao_janela(nome_janela)


@fixture
def coletar_situacao_janela_restaurada_test():
    nome_janela = 'Free Mouse Clicker'
    maximizar_janela(nome_janela)
    restaurar_janela(nome_janela)
    return coletar_situacao_janela(nome_janela)


@fixture
def minimizar_janela_test():
    nome_janela = 'Free Mouse Clicker'
    return minimizar_janela(nome_janela)


@fixture
def maximizar_janela_test():
    nome_janela = 'Free Mouse Clicker'
    return maximizar_janela(nome_janela)


@fixture
def restaurar_janela_test():
    nome_janela = 'Free Mouse Clicker'
    return restaurar_janela(nome_janela)


@fixture
def coletar_dados_selecao_test():
    caminho_campo = 'Free Mouse Clicker->combobox'
    return coletar_dados_selecao(caminho_campo)


@fixture
def coletar_dado_selecionado_test():
    caminho_campo = 'Free Mouse Clicker->combobox'
    return coletar_dado_selecionado(caminho_campo)


@fixture
def selecionar_em_campo_selecao_test():
    caminho_campo = 'Free Mouse Clicker->combobox'
    item = 'Single Click'
    return selecionar_em_campo_selecao(caminho_campo, item)


@fixture
def selecionar_menu_test():
    nome_janela = 'Sem título - Bloco de Notas'
    caminho_menu = '&Arquivo->Abrir'
    return selecionar_menu(nome_janela, caminho_menu)


@fixture
def fechar_janela_test():
    nome_janela = 'Sem título - Bloco de Notas'
    return fechar_janela(nome_janela)


@fixture
def caminho_pasta_exemplo():
    caminho = 'exemplo'
    return caminho


@fixture
def caminho_pasta_exemplo_2():
    caminho = 'exemplo/exemplo2'
    return caminho


@fixture
def caminho_pasta_exemplo_3():
    caminho = 'exemplo3'
    return caminho


@fixture
def caminho_pasta_exemplo_4():
    caminho = 'exemplo4'
    return caminho


@fixture
def caminho_pasta_exemplo_5():
    caminho = 'exemplo3/exemplo4'
    return caminho


@fixture
def contexto_manipulacao_pastas_vazias_excluir(caminho_pasta_exemplo):
    caminho = caminho_pasta_exemplo
    yield excluir_pasta(caminho)


@fixture
def contexto_manipulacao_pastas_vazias_criar(caminho_pasta_exemplo):
    caminho = caminho_pasta_exemplo
    criar_pasta(caminho)
    yield
    excluir_pasta(caminho, vazia=True)


@fixture
def contexto_manipulacao_pastas_cheias_criar(caminho_pasta_exemplo_2):
    caminho = caminho_pasta_exemplo_2
    yield criar_pasta(caminho)


@fixture
def contexto_manipulacao_pastas_cheias_excluir(caminho_pasta_exemplo_2):
    caminho = caminho_pasta_exemplo_2
    yield excluir_pasta(caminho, vazia=False)


@fixture
def contexto_manipulacao_pastas_renomear(
    caminho_pasta_exemplo_3, caminho_pasta_exemplo_4
):
    caminho3 = caminho_pasta_exemplo_3
    criar_pasta(caminho3)
    yield
    excluir_pasta(caminho_pasta_exemplo_4, vazia=True)


@fixture
def contexto_manipulacao_pastas_recortar(
    caminho_pasta_exemplo_3, caminho_pasta_exemplo_4, caminho_pasta_exemplo_5
):
    caminho3 = caminho_pasta_exemplo_3
    caminho4 = caminho_pasta_exemplo_4
    criar_pasta(caminho3)
    criar_pasta(caminho4)
    yield
    excluir_pasta(caminho_pasta_exemplo_3, vazia=False)
    excluir_pasta(caminho_pasta_exemplo_5)


@fixture
def contexto_manipulacao_pasta_copiar(
    caminho_pasta_exemplo, caminho_pasta_exemplo_3
):
    pasta = caminho_pasta_exemplo
    if not caminho_existente(pasta) == True:
        criar_pasta(pasta)
    caminho_destino = caminho_pasta_exemplo_3
    if not caminho_existente(caminho_destino) == True:
        criar_pasta(caminho_destino)
    yield
    excluir_pasta(pasta, vazia=True)
    excluir_pasta(caminho_destino, vazia=False)


@fixture
def contexto_manipulacao_pasta_mostar_arquivos(
    caminho_pasta_exemplo, caminho_raiz, arquivo_exemplo, arquivo_exemplo_2
):
    pasta = caminho_pasta_exemplo
    if not caminho_existente(pasta) == True:
        criar_pasta(pasta)
        criar_arquivo_texto(pasta + '/' + arquivo_exemplo)
        criar_arquivo_texto(pasta + '/' + arquivo_exemplo_2)
    yield
    excluir_pasta(pasta, vazia=False)


@fixture
def caminho_arquivo():
    caminho = './novo_arquivo_test.txt'
    return caminho


@fixture
def caminho_arquivo_2():
    caminho = 'tests/novo_arquivo_test.txt'
    return caminho


@fixture
def caminho_raiz():
    caminho = './'
    return caminho


@fixture
def arquivo_exemplo():
    arquivo = 'arquivo_test.txt'
    return arquivo


@fixture
def arquivo_exemplo_2():
    arquivo = 'arquivo_renomeado.txt'
    return arquivo


@fixture
def excluir_arquivo_test(caminho_arquivo):
    excluir_arquivo(caminho_arquivo)


@fixture
def contexto_manipulacao_arquivo_excluir(caminho_arquivo):
    caminho = caminho_arquivo
    if caminho_existente(caminho) == True:
        excluir_arquivo(caminho)
    yield
    excluir_arquivo(caminho)


@fixture
def contexto_manipulacao_arquivo_criar(caminho_arquivo):
    caminho = caminho_arquivo
    if not caminho_existente(caminho) == True:
        criar_arquivo_texto(caminho)
    yield
    excluir_arquivo(caminho)


@fixture
def contexto_manipulacao_arquivo_copiar(
    caminho_arquivo, caminho_pasta_exemplo
):
    arquivo = caminho_arquivo
    if not caminho_existente(arquivo) == True:
        criar_arquivo_texto(arquivo)
    if not caminho_existente(caminho_pasta_exemplo) == True:
        criar_pasta(caminho_pasta_exemplo)
    yield
    excluir_pasta(caminho=caminho_pasta_exemplo, vazia=False)
    excluir_arquivo(caminho=arquivo)


@fixture
def contexto_manipulacao_arquivo_criar_2(
    caminho_raiz, arquivo_exemplo, arquivo_exemplo_2
):
    caminho = caminho_raiz
    nome_arquivo = arquivo_exemplo
    novo_nome = arquivo_exemplo_2
    if not caminho_existente(nome_arquivo) == True:
        criar_arquivo_texto(caminho + nome_arquivo)
    yield
    excluir_arquivo(caminho + novo_nome)


@fixture
def contexto_manipulacao_arquivo_criar_3(caminho_arquivo, caminho_arquivo_2):
    caminho_atual = caminho_arquivo
    caminho_novo = caminho_arquivo_2
    if not caminho_existente(caminho_atual) == True:
        criar_arquivo_texto(caminho_atual)
    yield
    excluir_arquivo(caminho_novo)



"""Módulo para automação de aplicações desktop."""
# importa recursos do módulo pywinauto em nível global
from pywinauto import Application


__all__ = [
    'iniciar_app',
    'conectar_app',
    'retornar_janelas_disponiveis',
    'localizar_elemento',
    'capturar_texto',
    'digitar',
    'coletar_arvore_elementos',
    'localizar_diretorio_em_treeview',
    'clicar',
    'simular_clique',
    'simular_digitacao',
    'mover_mouse',
    'coletar_situacao_janela',
    'esta_visivel',
    'janela_existente',
    'esta_com_foco',
    'ativar_foco',
    'minimizar_janela',
    'maximizar_janela',
    'restaurar_janela',
    'coletar_dados_selecao',
    'coletar_dado_selecionado',
    'selecionar_em_campo_selecao',
    'selecionar_em_campo_lista',
    'selecionar_menu',
    'fechar_janela',
    'encerrar_app',
]

def _aplicacao(estilo_aplicacao: str = 'win32') -> Application:
    """Inicia e retorna um objeto do tipo Application da biblioteca pywinauto."""
    # define app como global
    global APP
    global ESTILO_APLICACAO


    ESTILO_APLICACAO = estilo_aplicacao

    # instancia o objeto application
    APP = Application(backend = ESTILO_APLICACAO)

    # retorna o objeto application instanciado
    return APP


def iniciar_app(
    executavel: str,
    estilo_aplicacao: str ='win32',
    esperar: tuple = (),
    inverter: bool = False,
    ocioso: bool = False,
) -> int:
    """Inicia e retorna um processo do sistema de um
    objeto do tipo Application com o caminho recebido."""
    # define app como global
    global APP
    global ESTILO_APLICACAO


    ESTILO_APLICACAO = estilo_aplicacao

    # instancia o objeto application
    APP = _aplicacao(estilo_aplicacao = ESTILO_APLICACAO)

    # inicia o processo de execução do aplicativo passado como parâmetro
    APP.start(
        cmd_line=executavel,
        wait_for_idle=ocioso,
    )

    esperar_por = tempo_espera = None
    # verifica se foi passado algum parâmetro para esperar, caso não:
    if esperar == ():
        # aguarda a inicialização da aplicação ficar pronta em até 10 segundos
        esperar_por = 'ready'
        tempo_espera = 10
    else:
        esperar_por, tempo_espera = esperar

    if inverter is False:
        # aguarda a inicialização da aplicação ficar na condição informada
        APP.window().wait(
            wait_for=esperar_por,
            timeout=tempo_espera,
            retry_interval=None,
        )
    else:
        # aguarda a inicialização da aplicação não ficar na condição informada
        APP.window().wait_not(
            wait_for_not=esperar_por,
            timeout=tempo_espera,
            retry_interval=None,
        )

    # coleta o PID da aplicação instanciada
    processo_app: int = APP.process

    # retorna o PID coletado
    return processo_app


def conectar_app(
    pid: int,
    tempo_espera: int = 60,
    estilo_aplicacao: str = 'win32',
) -> int:
    """Inicia e retorna um processo do sistema de um
    objeto do tipo Application com o caminho recebido."""
    # define app como global
    global APP
    global ESTILO_APLICACAO


    ESTILO_APLICACAO = estilo_aplicacao

    # instancia o objeto application
    APP = _aplicacao(estilo_aplicacao = ESTILO_APLICACAO)

    # inicia o processo de execução do aplicativo passado como parâmetro
    app_conectado: Application = _conectar_app(
        pid = pid,
        tempo_espera = tempo_espera,
        estilo_aplicacao = ESTILO_APLICACAO,
    )

    # coleta o PID da aplicação instanciada
    processo_app: int = app_conectado.process

    # retorna o PID coletado
    return processo_app


def _conectar_app(
    pid: int,
    tempo_espera: int = 60,
    estilo_aplicacao: str = 'win32',
) -> int:
    """Inicia e retorna um processo do sistema de um
    objeto do tipo Application com o caminho recebido."""
    # define app como global
    global APP
    global ESTILO_APLICACAO


    ESTILO_APLICACAO = estilo_aplicacao

    # instancia o objeto application
    APP = _aplicacao(estilo_aplicacao = ESTILO_APLICACAO)

    # inicia o processo de execução do aplicativo passado como parâmetro
    app_conectado: Application = APP.connect(
        process=pid,
        timeout=tempo_espera,
        backend=estilo_aplicacao,
    )

    # retorna o objeto Application atrelado ao PID informado
    return app_conectado


def retornar_janelas_disponiveis(
    pid: int,
    estilo_aplicacao = 'win32',
) -> str:
    """Retorna as janelas disponíveis em um
    objeto do tipo Application já em execução."""
    # importa app para o escopo da função
    global APP
    global ESTILO_APLICACAO


    ESTILO_APLICACAO = estilo_aplicacao

    # instancia o objeto application
    APP = _aplicacao(estilo_aplicacao = ESTILO_APLICACAO)


    # conecta a aplicação correspondente ao PID informado
    tempo_espera = 60
    app_interno: Application = _conectar_app(
        pid = pid,
        tempo_espera = tempo_espera,
        estilo_aplicacao = ESTILO_APLICACAO,
    )

    # coleta as janelas disponíveis
    lista_janelas = app_interno.windows()

    # instancia uma lista vazia
    lista_janelas_str = []
    # para cada janela na lista de janelas
    for janela in lista_janelas:
        # coleta e salva o nome da janela
        lista_janelas_str.append(janela.texts()[0])

    # retorna uma lista das janelas coletadas
    return lista_janelas_str


def localizar_elemento(
    caminho_campo: str,
    estatico: bool = True,
    estilo_aplicacao = 'win32',
) -> bool:
    """Retorna se o caminho de elementos informado existe
    no objeto do tipo Application sendo manipulado."""
    # importa app para o escopo da função
    global APP

    # inicializa APP para uma variável interna
    app_interno = _localizar_elemento(
        caminho_campo = caminho_campo,
        estatico = estatico,
    )

    return app_interno.exists()


def _localizar_elemento(
    caminho_campo: str,
    estatico: bool = True,
) -> Application:
    """Localiza e retorna um objeto do tipo Application
    percorrendo o caminho até o último o elemento."""
    # importa app para o escopo da função
    global APP

    # inicializa APP para uma variável interna
    app_interno = APP

    # trata o caminho da árvore de parantesco do app
    campo = caminho_campo.split('->')

    if estatico is False:
        campo[-1] = campo[-1] + 'Edit'

    # localiza o elemento até o final da árvore de parantesco do app:
    for index in range(len(campo)):
        # Se index for igual ao primeiro elemento
        if index == 0:
            # coleta um objeto application
            #   colocando o título como nome informado
            app_interno = app_interno.window(title=campo[0])
        else:
            # coleta o elemento informado e concatena 'Edit' no final
            app_interno = app_interno[campo[index]]

    return app_interno


def capturar_texto(caminho_campo: str, estatico: bool = True,) -> str:
    """Captura o texto de um elemento
    dentro de um objeto do tipo Application."""
    # localiza o elemento até o final da árvore de parantesco do app
    app_interno = _localizar_elemento(caminho_campo, estatico)

    # captura o texto do campo localizado
    valor_capturado: str = app_interno.texts()[0]

    # retorna o valor capturado
    return valor_capturado


def digitar(
    caminho_campo: str,
    valor: str,
    com_espaco: bool = True,
    digitacao_visivel: bool = True,
    com_nova_linha: bool = False,
) -> str:
    """Digita em um elemento dentro de um objeto do tipo Application."""
    # Define liberação para digitar
    estatico = False

    # localiza o elemento até o final da árvore de parantesco do app
    app_interno = _localizar_elemento(caminho_campo, estatico)

    # digita o valor no campo localizado
    app_interno.type_keys(
        keys = valor,
        with_spaces = com_espaco,
        set_foreground = digitacao_visivel,
        with_newlines = com_nova_linha,
    )

    # trata o valor capturado conforme o tipo do valor de entrada
    valor_retornado = str(capturar_texto(caminho_campo, estatico))

    # retorna o valor capturado e tratado
    return valor_retornado


def coletar_arvore_elementos(
    nome_janela: str,
    estatico: bool = True,
) -> list[str]:
    """Lista um elemento dentro de um objeto do
    tipo Application e retorna o valor coletado."""
    # importa recursos do módulo io
    import io
    # importa recursos do módulo Path
    from contextlib import redirect_stdout


    # localiza o elemento até o final da árvore de parantesco do app
    app_interno = _localizar_elemento(nome_janela, estatico)

    conteudoStdOut = io.StringIO()
    with redirect_stdout(conteudoStdOut):
        app_interno.print_control_identifiers()

    valor = conteudoStdOut.getvalue()
    valor_dividido = valor.split('\n')

    # retorna o valor capturado e tratado
    return valor_dividido


def localizar_diretorio_em_treeview(
    caminho_janela: str,
    caminho_diretorio: str,
    estatico: bool = True,
) -> bool:
    """Localiza um diretório, seguindo a árvore de diretórios informada,\
    dentro de um objeto TreeView do tipo Application."""
    try:
        # localiza e armazena o elemento conforme informado
        app_interno = _localizar_elemento(caminho_janela, estatico = estatico)

        # seleciona o caminho informado na janela do tipo TreeView
        app_interno.TreeView.get_item(caminho_diretorio).click()

        # clica em Ok para confirmar
        app_interno.OK.click()

        # retorna verdadeiro caso processo seja feito com sucesso
        return True
    except:
        return False


def clicar(caminho_campo: str) -> bool:
    """Clica em um elemento dentro de um objeto do tipo Application."""
    # localiza o elemento até o final da árvore de parantesco do app
    app_interno = _localizar_elemento(caminho_campo)

    # digita o valor no campo localizado
    app_interno.click()

    # retorna o valor capturado e tratado
    return True


def simular_clique(
    botao: str,
    eixo_x: int,
    eixo_y: int,
    tipo_clique: str = 'unico',
) -> bool:
    """Simula clique do mouse, performando o mouse real."""
    # importa recursos do módulo mouse
    from pywinauto.mouse import click, double_click

    if not botao.upper() in ['ESQUERDO', 'DIREITO']:
        raise ValueError('Informe um botão válido: esquerdo, direito.')

    if not tipo_clique.upper() in ['UNICO', 'DUPLO']:
        raise ValueError(
            'Tipo de clique inválido, escolha entre único e duplo.'
        )
    
    if (not isinstance(eixo_x, int)) or (not isinstance(eixo_y, int)):
        raise ValueError('Coordenadas precisam ser do tipo inteiro (int).')

    if botao.upper() == 'ESQUERDO':
        botao = 'left'
    else:
        botao = 'right'

    try:
        if tipo_clique.upper() == 'UNICO':
            click(button=botao, coords=(eixo_x, eixo_y))
        else:
            double_click(button=botao, coords=(eixo_x, eixo_y))

        return True
    except Exception:
        return False


def simular_digitacao(
    texto: str,
    com_espaco: bool = True,
    com_tab: bool = False,
    com_linha_nova: bool = False,
) -> bool:
    """Simula digitação do teclado, performando o teclado real."""
    # importa recursos do módulo keyboard
    from pywinauto.keyboard import send_keys

    if (not isinstance(com_espaco, bool)) \
    or (not isinstance(com_tab, bool)) \
    or (not isinstance(com_linha_nova, bool)):
        raise ValueError(
            """Informe os parâmetros com_espaco,
                com_tab e com_linha_nova com valor boleano"""
        )

    if (not isinstance(texto, str)):
        raise ValueError('Informe um texto do tipo string.')

    try:
        send_keys(
            keys=texto,
            with_spaces=com_espaco,
            with_tabs=com_tab,
            with_newlines=com_linha_nova,
        )

        return True
    except:
        return False


def mover_mouse(eixo_x: int, eixo_y: int) -> bool:
    # importa recursos do módulo mouse
    from pywinauto.mouse import move

    if (not isinstance(eixo_x, int)) \
    or (not isinstance(eixo_y, int)):
        raise ValueError('Coordenadas precisam ser do tipo inteiro (int).')

    try:
        move(coords=(eixo_x, eixo_y))

        return True
    except:
        return False


def coletar_situacao_janela(nome_janela: str) -> str:
    """Coleta a situação do estado atual de uma
    janela de um objeto do tipo Application."""
    # importa app para o escopo da função
    global APP

    # inicializa APP para uma variável interna
    app_interno = APP

    situacao = ''
    # coleta a situacao atual da janela
    situacao_temp = (
        app_interno
            .window(title = nome_janela)
            .get_show_state()
    )

    # 1 - Normal
    # 2 - Minimizado
    # 3 - Maximizado
    # Caso não encontre as situações normal, ninimizado e
    #   maximizado, define um valor padrão.
    if situacao_temp == 1:
        situacao = 'normal'
    elif situacao_temp == 2:
        situacao = 'minimizado'
    elif situacao_temp == 3:
        situacao = 'maximizado'
    else:
        situacao = 'não identificado'

    # retorna a situação da janela
    return situacao


def esta_visivel(nome_janela: str) -> str:
    """Verifica se a janela de um objeto do tipo Application está visível."""
    # coleta a situação atual da janela
    situacao = coletar_situacao_janela(nome_janela)

    # define visível para situação 'maximizado' ou 'normal'
    if situacao == 'maximizado' or situacao == 'normal':
        situacao = 'visivel'
    # define não visível para situação 'minimizado'
    elif situacao == 'minimizado':
        situacao = 'não visível'
    # Caso não encontre as situações normal, ninimizado e maximizado
    else:
        # define um valor padrão
        situacao = 'não identificado'

    # retorna a situação da janela
    return situacao


def janela_existente(pid, nome_janela) -> bool:
    """Verifica se a janela de um objeto do tipo Application está visível."""
    # coleta a situação atual da janela
    lista_janelas = retornar_janelas_disponiveis(pid)

    # verifica se o nome da janela informada corresponde à alguma janela na lista
    for janela in lista_janelas:
        # caso o nome da janela seja o mesmo da janela atual da lista
        if janela == nome_janela:
            # retorna True
            return True

    # retorna False caso nenhuma janela tenha correspondido
    return False


def esta_com_foco(nome_janela: str) -> bool:
    """Verifica se a janela de um objeto do tipo Application está com foco."""
    # importa app para o escopo da função
    global APP

    # inicializa APP para uma variável interna
    app_interno = APP

    # retorna a situacao atual de foco da janela
    return app_interno.window(title = nome_janela).has_focus()


def ativar_foco(nome_janela: str) -> bool:
    """Ativa a janela de um objeto do tipo Application deixando-a com foco."""
    # importa app para o escopo da função
    global APP

    try:
        # inicializa APP para uma variável interna
        app_interno = APP

        # ativa a janela informada
        app_interno.window(title = nome_janela).set_focus()

        # retorna verdadeiro confirmando a execução da ação
        return True
    except:
        return False


def minimizar_janela(nome_janela: str) -> bool:
    """Miniminiza a janela de um objeto do tipo Application."""
    # importa app para o escopo da função
    global APP

    try:
        # inicializa APP para uma variável interna
        app_interno = APP

        # miniminiza a janela informada
        app_interno.window(title = nome_janela).minimize()

        # retorna verdadeiro confirmando a execução da ação
        return True
    except:
        return False


def maximizar_janela(nome_janela: str) -> bool:
    """Maximiza a janela de um objeto do tipo Application."""
    # importa app para o escopo da função
    global APP

    try:
        # inicializa APP para uma variável interna
        app_interno = APP

        # maximiza a janela informada
        app_interno.window(title = nome_janela).maximize()

        # retorna verdadeiro confirmando a execução da ação
        return True
    except:
        return False


def restaurar_janela(nome_janela: str) -> bool:
    """Miniminiza a janela de um objeto do tipo Application."""
    # importa app para o escopo da função
    global APP

    try:
        # inicializa APP para uma variável interna
        app_interno = APP

        # restaura a janela informada
        app_interno.window(title = nome_janela).restore()

        # retorna verdadeiro confirmando a execução da ação
        return True
    except:
        return True


def coletar_dados_selecao(caminho_campo: str) -> str:
    """Coleta dados disponíveis para seleção em um
    elemento de seleção em um objeto do tipo Application."""
    # define estático como falso para trabalhar com elemento dinâmico
    estatico = False

    # localiza o elemento até o final da árvore de parantesco do app
    app_interno = _localizar_elemento(caminho_campo, estatico)

    # captura o texto do campo localizado
    valor_capturado: str = app_interno.item_texts()

    # retorna o valor capturado
    return valor_capturado


def coletar_dado_selecionado(caminho_campo: str) -> str:
    """Coleta dado já selecionado em um elemento
    de seleção em um objeto do tipo Application."""
    # define estático como falso para trabalhar com elemento dinâmico
    estatico = False

    # localiza o elemento até o final da árvore de parantesco do app
    app_interno = _localizar_elemento(caminho_campo, estatico)

    # captura o texto do campo localizado
    valor_capturado: str = app_interno.selected_text()

    # retorna o valor capturado
    return valor_capturado


def selecionar_em_campo_selecao(caminho_campo: str, item: str) -> str:
    """Seleciona um dado em um elemento de
    seleção em um objeto do tipo Application."""
    # define estático como falso para trabalhar com elemento dinâmico
    estatico = False

    # localiza o elemento até o final da árvore de parantesco do app
    app_interno = _localizar_elemento(caminho_campo, estatico)

    # seleciona o item informado
    app_interno.select(item)

    # captura o texto do campo localizado
    valor_capturado = coletar_dado_selecionado(caminho_campo)

    # retorna o valor capturado
    return valor_capturado


def selecionar_em_campo_lista(
    caminho_campo: str,
    item: str,
    estatico = False,
) -> str:
    """Seleciona um dado em um elemento de
    lista em um objeto do tipo Application."""

    # localiza o elemento até o final da árvore de parantesco do app
    app_interno = _localizar_elemento(caminho_campo, estatico)

    # seleciona o item informado
    app_interno.select(item)

    # captura o índice do ítem selecionado
    indice_selecionado = app_interno.selected_indices()

    # retorna o valor capturado
    return indice_selecionado


def selecionar_menu(nome_janela: str, caminho_menu: str) -> bool:
    """Seleciona um item de menu conforme o caminho
    informado em um objeto do tipo Application."""
    # importa app para o escopo da função
    global APP

    try:
        # inicializa APP para uma variável interna
        app_interno = APP

        # percorre e clica no menu informado
        app_interno.window(title = nome_janela).menu_select(caminho_menu)

        # retorna verdadeiro confirmando a execução da ação
        return True
    except:
        return False


def fechar_janela(nome_janela: str) -> bool:
    """Encerra uma janela de um objeto do tipo
    Application com o caminho recebido."""
    # importa app para o escopo da função
    global APP

    # inicializa APP para uma variável interna
    app_interno = APP

    # fecha a janela informada
    app_interno.window(title = nome_janela).close()

    # retorna verdadeiro confirmando a execução da ação
    return True


def encerrar_app(
    pid: int,
    forcar: bool = False,
    tempo_espera: int = 60,
) -> bool:
    """Encerra e retorna um processo do sistema de um
    objeto do tipo Application com o caminho recebido."""
    # importa app para o escopo da função
    global APP

    # conecta a aplicação correspondente ao PID informado
    app_interno: Application = _conectar_app(
        pid = pid,
        tempo_espera = tempo_espera,
        estilo_aplicacao = ESTILO_APLICACAO,
    )

    # encerra o aplicativo em execução
    app_interno.kill(soft = not forcar)

    # retorna o objeto application com o processo encerrado
    return True



"""Módulo para facilidades no manuseio de recursos comuns no desenvolvimento."""


__all__ = [
    'abrir_arquivo_em_bytes',
    'abrir_arquivo_excel',
    'abrir_arquivo_pdf',
    'abrir_arquivo_texto',
    'abrir_arquivo_word',
    'adicionar_ao_zip',
    'alterar_arquivo_texto',
    'caminho_existente',
    'cls',
    'coletar_arvore_caminho',
    'coletar_caminho_absoluto',
    'coletar_extensao_arquivo',
    'coletar_idioma_so',
    'coletar_nome_arquivo',
    'coletar_nome_guias_arquivo_excel',
    'coletar_pid',
    'compactar',
    'converter_pdf_em_imagem',
    'copiar_arquivo',
    'copiar_pasta',
    'criar_arquivo_texto',
    'criar_pasta',
    'descompactar',
    'escrever_em_arquivo',
    'excluir_arquivo',
    'excluir_pasta',
    'extrair_texto_ocr',
    'finalizar_processo',
    'gravar_log_em_arquivo',
    'janela_dialogo',
    'ler_variavel_ambiente',
    'logar',
    'pasta_esta_vazia',
    'processo_existente',
    'recortar',
    'remover_acentos',
    'renomear',
    'retornar_arquivos_em_pasta',
    'retornar_data_hora_atual',
]


def abrir_arquivo_em_bytes(caminho):
    """Abre em bytes um arquivo de texto no caminho informado."""
    # importa recursos do módulo Path
    from pathlib import Path

    # abre um arquivo de texto e coleta o conteúdo em bytes
    arquivo = Path(caminho).read_bytes()

    # retorna o conteúdo do arquivo
    return arquivo


def abrir_arquivo_excel(
    arquivo_excel: str,
    guia: str = '',
    manter_macro: bool = True,
    manter_links: bool = True,
):
    """Abre um arquivo de Excel no caminho informado."""

    # importa recursos do módulo openpyxl
    import xlrd
    from openpyxl import load_workbook

    # trata o caminho com o objeto Path
    caminho_excel = coletar_caminho_absoluto(arquivo_excel)
    extensao_arquivo_excel = coletar_extensao_arquivo(caminho_excel)

    # define um valor padrão e inicial à lista
    tabela_excel = []

    if extensao_arquivo_excel[0].upper() == '.XLS':
        # abre um arquivo de Excel e coleta o conteúdo
        conteudo_excel = xlrd.open_workbook(
            caminho_excel,
        )

        # seleciona a guia à trabalhar
        if guia == '':
            aba_ativa = conteudo_excel.sheet_by_index(0)
        else:
            aba_ativa = conteudo_excel.sheet_by_name(guia)

        # para cada linha do conteúdo coletado
        for indice_linha in range(aba_ativa.nrows):
            # adiciona a linha na tabela
            tabela_excel.append(aba_ativa.row_values(indice_linha))
    else:
        # abre um arquivo de Excel e coleta o conteúdo
        conteudo_excel = load_workbook(
            caminho_excel,
            keep_vba=manter_macro,
            keep_links=manter_links,
        )

        # seleciona a guia à trabalhar
        if guia == '':
            aba_ativa = conteudo_excel.active
        else:
            aba_ativa = conteudo_excel[guia]

        # para cada linha do conteúdo coletado
        for linhas in aba_ativa.values:
            # define um valor padrão e inicial à lista
            linha = []

            # para cada valor na célula da linha
            for celula in linhas:
                # adiciona o valor na linha
                linha.append(celula)

            # adiciona a linha na tabela
            tabela_excel.append(linha)

    # retorna o conteúdo da tabela
    return tabela_excel


def abrir_arquivo_pdf(
    arquivo_pdf: str,
    senha_pdf: str or None = None,
    paginacao: int or tuple[int] = 0,
    orientacao: int = 0,
):
    """Abre um arquivo de word no caminho informado."""

    # importa recursos do módulo PyPDF2
    from PyPDF2 import PdfReader

    # trata o caminho com o objeto Path
    caminho_pdf = coletar_caminho_absoluto(arquivo_pdf)

    # abre um arquivo de PDF e coleta o conteúdo
    conteudo_pdf = PdfReader(
        stream=caminho_pdf,
        password=senha_pdf,
        strict=False,
    )

    # define um valor padrão e inicial à lista
    lista_paginacao = []

    # caso o tipo do parâmetro 'paginacao' seja do tipo int
    if isinstance(paginacao, int):
        # transforma 'paginacao' em tupla
        paginacao = (paginacao,)

    # se paginacao for igual à 0
    if 0 in paginacao:
        # adiciona todas as páginas para a lista de paginações
        lista_paginacao = conteudo_pdf.pages
    else:
        # para cada valor do índice de 'paginação'
        for indice in paginacao:
            # caso índice seja do tipo int
            if isinstance(indice, int) is True:
                # ajusta o índice ao padrão de índice em listas da linguagem
                indice = indice - 1

                # adiciona a paginação solicitada à lista de paginações
                lista_paginacao.append(conteudo_pdf.getPage(indice))
            # caso índice não seja do tipo int
            else:
                # levanta exceção de tipo incorreto
                raise TypeError(
                    'Parâmetro ``paginacao`` aceita somente ítens numéricos (int).'
                )

    # define um valor padrão e inicial à lista
    lista_texto_pdf = []

    # para cada página do conteúdo coletado
    for pagina in lista_paginacao:
        # adiciona o valor na linha separando por páginas [n] e linhas [n][n]
        lista_texto_pdf.append(
            pagina.extract_text(orientations=orientacao).splitlines()
        )

    # retorna o conteúdo coletado em lista
    return lista_texto_pdf


def abrir_arquivo_texto(caminho, encoding='utf8'):
    """Abre um arquivo de texto no caminho informado."""
    # importa recursos do módulo Path
    from pathlib import Path

    # abre um arquivo de texto e coleta o conteúdo
    arquivo = Path(caminho).read_text(encoding=encoding)

    # retorna o conteúdo do arquivo
    return arquivo


def adicionar_ao_zip(
    caminho: str,
    arquivo_destino: str,
    recursivo: bool = False,
):
    """Adiciona um caminho à um arquivo zip informado."""
    # importa recursos do módulo zipfile
    from zipfile import ZipFile

    if recursivo is True:
        filtro = './**/*'
        lista_caminhos = retornar_arquivos_em_pasta(
            caminho=caminho,
            filtro=filtro,
        )
    else:
        lista_caminhos = [
            caminho,
        ]

    with ZipFile(arquivo_destino, 'a') as objeto_zip:
        for arquivo in lista_caminhos:
            if len(lista_caminhos) == 1:
                caminho = coletar_arvore_caminho(lista_caminhos[0])

            caminho_interno_zip = arquivo.replace(caminho, '')
            objeto_zip.write(
                filename=arquivo,
                arcname=caminho_interno_zip,
            )

    return True


def alterar_arquivo_texto(
    caminho,
    linha_atual,
    linha_alterada,
    multilinhas=False,
    encoding_entrada='utf8',
    encoding_saida='utf8',
):
    """Abre um arquivo de texto no caminho informado."""
    # importa recursos do módulo Path
    from pathlib import Path

    # trata o caminho com o objeto Path
    arquivo = Path(caminho)
    # abre um arquivo de texto e coleta o conteúdo
    conteudo = arquivo.read_text(encoding=encoding_entrada)
    # divide o conteúdo por linhas ('\n')
    conteudo = conteudo.splitlines()

    # define um valor padrão para a variável
    validacao_multilinhas = True

    # para cada linha do arquivo .txt aberto
    for linha_arquivo in range(len(conteudo)):
        # se for a primeira linha, define o modo de escrita
        if linha_arquivo == 0:
            modo = 'w'
        # se não for a primeira linha, define o modo de concatenação
        else:
            modo = 'a'

        # abre um arquivo de texto para alteração
        with open(arquivo, modo, encoding=encoding_saida) as arquivo_aberto:
            # caso seja definido multiplas linhas:
            if validacao_multilinhas is True:
                # se a linha atual corresponder ao conteúdo passado no parâmetro
                if conteudo[linha_arquivo].__contains__(linha_atual):
                    # substitui a linha atual pelo conteúdo passada no parâmetro
                    conteudo[linha_arquivo] = linha_atual.replace(
                        linha_atual,
                        linha_alterada,
                    )

                    # se não for definido multiplas linha no parâmetro
                    if multilinhas is False:
                        # anula nova entrada desse bloco
                        validacao_multilinhas = False

            # escreve o conteúdo no arquivo
            arquivo_aberto.write(conteudo[linha_arquivo] + '\n')

    # retorna o conteúdo do arquivo
    return conteudo


def caminho_existente(caminho):
    """Verifica se um arquivo no caminho informado existe."""
    # importa recursos do módulo Path
    from pathlib import Path

    # coleta o caminho absoluto do caminho informado
    caminho = coletar_caminho_absoluto(caminho)

    # verifica e retorna se o arquivo existe.
    #   True caso exista e False se não existir.
    return Path(caminho).exists()


def cls():
    """Limpa a visualização do terminal."""
    import os

    os.system('cls')


def coletar_arvore_caminho(caminho):
    """Retorna os arquivos existentes em um diretório se seus
    respectivos sub-diretórios segundo o caminho informado."""
    # importa recursos do módulo Path
    from pathlib import Path

    # coleta o caminho informado no padrão do objeto Path
    caminho_interno = coletar_caminho_absoluto(caminho)

    # coleta o caminho informado no padrão do objeto Path
    caminho_interno = Path(caminho_interno)

    # coleta a árvore do caminho informado
    arvore_caminho = str(caminho_interno.parent)

    # retorna o caminho absoluto coletado
    return arvore_caminho


def coletar_caminho_absoluto(caminho):
    """Retorna os arquivos existentes em um diretório se seus
    respectivos sub-diretórios segundo o caminho informado."""
    # importa recursos do módulo Path
    from pathlib import Path

    # coleta o caminho informado no padrão do objeto Path
    caminho_interno = Path(caminho)

    # coleta o caminho absoluto do caminho
    caminho_absoluto = str(caminho_interno.absolute())

    # retorna o caminho absoluto coletado
    return caminho_absoluto


def coletar_extensao_arquivo(caminho):
    """Coleta a extensão de um arquivo no caminho informado."""
    # importa recursos do módulo Path
    from pathlib import Path

    # coleta a extensão do arquivo
    arquivo = Path(caminho).suffixes

    # retorna a extensão coletada
    return arquivo


def coletar_idioma_so():
    """Coleta o idioma atual do sistema operacional."""
    # importa recursos do módulo ctypes
    import ctypes

    # importa recursos do módulo locale
    import locale

    # coleta as informações do kernel do Windows
    windows_dll = ctypes.windll.kernel32

    # coleta o valor do idioma local do sistema no padrão de ID numÃ©rico
    windows_dll.GetUserDefaultUILanguage()

    # coleta o valor do idioma local do sistema no padrão de escrita abreviada
    idioma = locale.windows_locale[windows_dll.GetUserDefaultUILanguage()]

    # retorna o valor de idioma coletado
    return idioma


def coletar_nome_arquivo(caminho):
    """Coleta o nome de um arquivo no caminho informado."""
    # importa recursos do módulo Path
    from pathlib import Path

    # coleta o nome do arquivo informado
    arquivo = Path(caminho).stem

    # retorna o nome do arquivo
    return arquivo


def coletar_nome_guias_arquivo_excel(arquivo_excel):
    """Coleta as guias existentes no arquivo Excel informado."""

    # importa recursos do módulo openpyxl
    from openpyxl import load_workbook

    # trata o caminho com o objeto Path
    caminho_excel = coletar_caminho_absoluto(arquivo_excel)

    # abre um arquivo de Excel e coleta o conteúdo
    conteudo_excel = load_workbook(caminho_excel)

    # define um valor padrão e inicial à lista
    lista_guias = []

    # coleta a lista de guias que o arquivo contém
    lista_guias = conteudo_excel.sheetnames

    # retorna a lista coletada
    return lista_guias


def coletar_pid(nome_processo):
    """Coleta o idioma atual do sistema operacional."""
    # importa recursos do módulo os
    import psutil

    # instancia uma lista vazia
    listaProcessos = []
    # para cada processo na lista de processos
    for processo in psutil.process_iter():
        # tenta executar a ação
        try:
            # coleta o PID, o nome, o tempo de início do processo
            informacao_processo = processo.as_dict(
                attrs=['pid', 'name', 'create_time']
            )

            # se existir um processo com o mesmo nome informado
            if nome_processo.lower() in informacao_processo['name'].lower():
                # salva o nome do processo
                listaProcessos.append(informacao_processo)
        # para a lista de erros informados
        except (
            psutil.NoSuchProcess,
            psutil.AccessDenied,
            psutil.ZombieProcess,
        ):
            # ignora os erros
            ...

    # retorna uma lista de dicionários com o nome do processo coletado
    return listaProcessos


def compactar(
    caminho: str,
    arquivo_destino: str,
    modo: str = 'w',
):
    """Compacta um caminho para o arquivo zip informado."""
    # importa recursos do módulo zipfile
    from zipfile import ZipFile

    lista_caminhos = retornar_arquivos_em_pasta(
        caminho=caminho,
        filtro='./**/*',
    )

    with ZipFile(arquivo_destino, modo) as objeto_zip:
        for arquivo in lista_caminhos:
            caminho_interno_zip = arquivo.replace(caminho, '')
            objeto_zip.write(
                filename=arquivo,
                arcname=caminho_interno_zip,
            )

    return True


def converter_pdf_em_imagem(
    arquivo_pdf: str,
    caminho_saida: str,
    alpha: bool = False,
    zoom: float = 1,
    orientacao: int = 0,
):
    """Converte cada página de um arquivo PDF em imagem."""
    # importa recursos do módulo Path
    from pathlib import Path

    # importa recursos do módulo fitz
    import fitz

    try:
        # trata os caminhos com o objeto Path
        caminho_pdf = coletar_caminho_absoluto(arquivo_pdf)
        caminho_saida_img = coletar_caminho_absoluto(caminho_saida)

        # abre um arquivo de PDF e coleta o conteúdo
        conteudo_pdf = fitz.open(caminho_pdf)

        # para cada página
        for indice in range(conteudo_pdf.page_count):
            # coleta a página atual
            pagina = conteudo_pdf[indice]
            # coleta a rotacao
            rotacao = orientacao
            # coleta o zoom do eixo X e eixo Y
            zoom_x = zoom_y = zoom
            # trata o arquivo de saída
            arquivo_img = (
                Path(caminho_saida_img) / f'arquivo_{str(indice+1)}.png'
            )

            # coleta a matriz da página, combinando o zoom e a rotação
            matriciado = fitz.Matrix(zoom_x, zoom_y).prerotate(rotacao)
            # converte a matriz da página em um mapa
            #   de píxel de imagem adicionando fundo
            mapa_pixel = pagina.get_pixmap(matrix=matriciado, alpha=alpha)

            # salva o mapa de píxel em um arquivo de imagem
            mapa_pixel.save(arquivo_img)

        # retorna True caso a operação tenha sucesso
        return True
    except Exception as erro:
        # retorna False caso a operação tenha sucesso
        raise erro


def copiar_arquivo(arquivo, caminho_destino):
    """Copia um arquivo de um caminho para
    outro caminho conforme informado."""

    # coleta o caminho absoluto do arquivo
    arquivo = coletar_caminho_absoluto(arquivo)

    # importa recursos do módulo shutil
    from shutil import copy2

    # copia o arquivo para a pasta de destino informado
    caminho_destino = copy2(arquivo, caminho_destino)

    # retorna o caminho de destino
    return str(caminho_destino)


def copiar_pasta(pasta, caminho_destino):
    """Copia uma pasta de um caminho para outro caminho conforme informado."""
    # importa recursos do módulo Path
    from pathlib import Path

    # trata o caminho de destino com o objeto Path
    caminho_origem = coletar_caminho_absoluto(pasta)
    caminho_destino = coletar_caminho_absoluto(caminho_destino)

    # importa recursos do módulo shutil
    from shutil import copy, copy2, copytree

    Path(caminho_destino)

    # copia a pasta para o destino informado
    copytree(
        caminho_origem,
        Path(caminho_destino) / Path(caminho_origem).stem,
    )

    # retorna o caminho de destino com a pasta copiada
    return caminho_destino


def criar_arquivo_texto(
    caminho,
    dado='',
    encoding='utf8',
    em_bytes: bool = False,
):
    # importa recursos do módulo Path
    """Cria um arquivo de texto no caminho informado."""
    from pathlib import Path

    # caso em_bytes não for verdadeiro
    if em_bytes is False:
        # escreve em um arquivo de texto o conteúdo informado
        Path(caminho).write_text(encoding=encoding, data=dado)
    # caso em_bytes for verdadeiro
    else:
        # escreve em um arquivo de texto o conteúdo informado em bytes
        Path(caminho).write_bytes(data=dado)

    # retorna True caso a operação tenha concluída com sucesso
    return True


def criar_pasta(caminho):
    """Cria pasta com caminho e nome informado."""
    # importa recursos do módulo Path
    from pathlib import Path

    # trata o caminho com o objeto Path
    caminho_interno = Path(caminho)

    # cria a pasta informada, caso necessário cria
    #   a hierarquia anterior à última pasta
    caminho_interno.mkdir(parents=True)

    # retorna True caso a operação tenha concluída com sucesso
    return True


def descompactar(arquivo, caminho_destino, senha_arquivo=None):
    """Descompacta um arquivo para o caminho informado."""
    # importa recursos do módulo zipfile
    from zipfile import ZipFile

    with ZipFile(file=arquivo, mode='r') as objeto_zip:
        objeto_zip.extractall(path=caminho_destino, pwd=senha_arquivo)


def escrever_em_arquivo(
    arquivo,
    conteudo,
    modo,
    encoding='utf8',
    nova_linha=None,
):
    """salva o conteúdo informado em um arquivo de texto também informado."""
    from pathlib import Path

    caminho_arquivo = Path(arquivo)
    caminho_arquivo = coletar_caminho_absoluto(caminho_arquivo)

    if (nova_linha is None) or (
        (nova_linha is not None) and (nova_linha not in ['\r', '\n', '\r\n'])
    ):
        nova_linha = ''

    # abre o arquivo definindo o modo de edição e o encoding
    with open(
        caminho_arquivo,
        modo,
        encoding=encoding,
    ) as arquivo:
        # escreve efetivamente o conteúdo no arquivo
        arquivo.write(conteudo + nova_linha)

    # fecha o arquivo
    arquivo.close()


def excluir_arquivo(caminho):
    """Exclui um arquivo no caminho informado."""
    # importa recursos do módulo Path
    from pathlib import Path

    caminho = coletar_caminho_absoluto(caminho)

    # exclui o arquivo informado
    Path(caminho).unlink()

    # retorna True caso a operação tenha concluída com sucesso
    return True


def excluir_pasta(caminho, vazia: bool = True):
    """Exclui pasta no caminho informado. Caso a pasta não esteja vazia,
    informe explicitamente no parâmetro 'vazia'."""

    caminho_interno = coletar_caminho_absoluto(caminho)

    # Se a pasta estiver vazia
    if vazia is True:
        # importa recursos do módulo Path
        from pathlib import Path

        # exclui a pasta informada
        Path(caminho_interno).rmdir()

        # retorna True caso a operação tenha concluída com sucesso
        return True
    # Se a pasta não estiver vazia
    else:
        # importa recursos do módulo rmtree
        from shutil import rmtree

        # exclui a pasta informada e o conteúdo contido nela
        rmtree(caminho_interno)

        # retorna True caso a operação tenha concluída com sucesso
        return True


def extrair_texto_ocr(arquivo, linguagem, encoding='utf8'):
    """Extrai texto de arquivo de imagem usando OCR."""
    # importa recursos do módulo subprocess
    import subprocess

    # abre um arquivo de texto e coleta o conteúdo em bytes
    caminho_arquivo = coletar_caminho_absoluto(arquivo)

    # coleta o texto da imagem usando Pytesseract OCR
    texto_extraido = subprocess.run(
        ('pytesseract', '-l', linguagem, caminho_arquivo),
        stdout=subprocess.PIPE,
        encoding=encoding,
    )

    # retorna o texto coletado
    return texto_extraido.stdout


def finalizar_processo(pid: int):
    """Coleta o idioma atual do sistema operacional."""
    # importa recursos do módulo os
    import psutil

    # instancia um dicionário vazio
    listaProcessos = {}
    # para cada processo na lista de processos
    for processo in psutil.process_iter():
        # tenta executar a ação
        try:
            # coleta o PID, o nome, o tempo de início do processo
            informacao_processo = processo.as_dict(
                attrs=['pid', 'name', 'create_time']
            )

            # se existir um processo com o mesmo nome informado
            if pid == informacao_processo['pid']:
                # encerra o processo informado
                processo.kill()

                # retorna true
                return True
        # para a lista de erros informados
        except (
            psutil.NoSuchProcess,
            psutil.AccessDenied,
            psutil.ZombieProcess,
        ):
            # ignora os erros
            ...

    # retorna um dicionário com o nome do processo coletado
    return False


def gravar_log_em_arquivo(
    arquivo,
    conteudo,
    modo,
    encoding='utf8',
    delimitador=';',
    nova_linha='\r\n',
):
    """salva o conteúdo informado em um arquivo de texto também informado."""

    # transforma todos os argumentos em lista
    if (not isinstance(conteudo, list)) or (not isinstance(conteudo, tuple)):
        conteudo = list(conteudo)

    # define a variavel
    conteudo = delimitador.join(conteudo)

    # abre o arquivo definindo o modo de edição e o encoding
    escrever_em_arquivo(
        arquivo=arquivo,
        conteudo=conteudo,
        modo=modo,
        encoding=encoding,
        nova_linha=nova_linha,
    )


def janela_dialogo(titulo: str, texto: str, estilo: int = 1):
    """Exibe uma janela de mensagem na tela."""
    # importa recursos do módulo ctypes
    import ctypes

    # cria o objeto de janela conforme os parâmentros informados
    caixa = ctypes.windll.user32.MessageBoxW(0, texto, titulo, estilo)

    # retorna o objeto
    return caixa


def ler_variavel_ambiente(
    arquivo_config='config.ini',
    nome_bloco_config='padrao',
    nome_variavel=None,
    variavel_sistema: bool = False,
    encoding='utf8',
):
    """Lê uma variável de ambiente,
    tanto de um arquivo quanto direto do sistema."""
    # importa recursos do módulo os
    import os

    # importa recursos do módulo ConfigParser
    from configparser import ConfigParser

    # se não for variável de sistema
    if not variavel_sistema is True:
        # instancia o objeto de configuração
        config = ConfigParser()
        # Lê o arquivo de configuração
        config.read(arquivo_config, encoding=encoding)

        # se o nome da variável de ambiente foi informada
        if nome_variavel is not None:
            # coleta o dado da variável de ambiente informado
            bloco = dict(config[nome_bloco_config])
            # retorna o valor coletado
            return bloco[nome_variavel]
        # se o nome da variável de ambiente não foi informada
        else:
            # retorna o todos os dados no
            #   bloco de variáveis de ambiente informado
            return dict(config[nome_bloco_config])
    # se for variável de sistema
    else:
        # retorna o valor da variável de sistema solicitado
        return os.environ.get(nome_variavel)


def logar(
    mensagem,
    nivel,
    arquivo=None,
    modo=None,
    encoding=None,
    formatacao=None,
    handlers=None,
):
    """Formata e retorna uma string como log.
    Será exibido sempre o ní­vel em primeira posição."""
    # importa recursos do módulo logging
    from logging import (
        CRITICAL,
        DEBUG,
        ERROR,
        INFO,
        WARNING,
        basicConfig,
        critical,
        debug,
        error,
        info,
        warning,
    )

    # define um ní­vel de log
    nivel = nivel.upper()

    # define configurações básicas de log
    basicConfig(
        level=nivel,
        filename=arquivo,
        filemode=modo,
        encoding=encoding,
        format=formatacao,
        handlers=handlers,
    )

    # executa comando de logging conforme o nível:
    if nivel == 'DEBUG':
        debug(mensagem)
    elif nivel == 'INFO':
        info(mensagem)
    elif nivel == 'WARNING':
        warning(mensagem)
    elif nivel == 'ERROR':
        error(mensagem)
    elif nivel == 'CRITICAL':
        critical(mensagem)
    # caso o nível não corresponder aos ní­veis padrões de logging
    else:
        # retorna mensagem de parâmetro inválido
        return 'Parâmetro nível inválido. Por favor, informe-o corretamente.'

    # retorna a mensagem e o nível
    return (nivel, mensagem)


def pasta_esta_vazia(caminho):
    """Verifica se uma pasta no caminho informado está vazia."""
    # importa recursos do módulo Path
    from pathlib import Path

    caminho = coletar_caminho_absoluto(caminho)
    caminho = Path(caminho)

    # se existir o caminho informado
    if caminho_existente(caminho):
        # coleta de forma recursiva o conteúdo
        #   contido no caminho informado caso existir
        lista_arquivos_pastas = list(caminho.glob('**/*'))

        # se não existir conteúdo no caminho informado
        if len(lista_arquivos_pastas) == 0:
            # retorna True informando que a pasta está vazia
            return True

    # retorna False informando que a pasta não está vazia
    return False


def processo_existente(nome_processo):
    """Coleta o idioma atual do sistema operacional."""
    # importa recursos do módulo psutil
    import psutil

    # para cada processo na lista de processos
    for processo in psutil.process_iter():
        # tenta executar a ação
        try:
            # verifica se o nome do processo corresponde ao nome informado
            if nome_processo.lower() in processo.name().lower():
                # caso exista retorna True
                return True
        # para a lista de erros informados
        except (
            psutil.NoSuchProcess,
            psutil.AccessDenied,
            psutil.ZombieProcess,
        ):
            # ignora os erros
            ...
    # retorna False caso não encontre processo com o nome informado
    return False


def recortar(caminho_atual, caminho_novo):
    """Recorta um arquivo ou pasta de um caminho
    e cola em outro caminho conforme informado."""
    # importa recursos do módulo Path
    from pathlib import Path

    # trata o caminho atual com o objeto Path
    caminho_atual = Path(caminho_atual)
    # trata o caminho novo com o objeto Path
    caminho_novo = Path(caminho_novo)

    final_caminho = caminho_atual.stem + caminho_atual.suffix

    # modifica o nome conforme informado
    caminho_novo = caminho_atual.rename(caminho_novo / final_caminho)

    # retorna o objeto Path com o caminho novo
    return caminho_novo


def remover_acentos(
    texto,
    normalizacao='NFKD',
):
    """Cria pasta com caminho e nome informado."""
    # importa recursos do módulo unicodedata
    # importa recursos do módulo re
    import re
    import unicodedata

    # separa os caracteres comuns dos especiais
    # '''
    separacao_acentos = unicodedata.normalize(
        normalizacao,
        texto,
    )

    texto_tratado = ''
    for caractere in separacao_acentos:
        if ord(caractere) < 128:
            texto_tratado += ''.join([caractere])
        else:
            ...

    # remove os caracteres especiais
    texto_limpo = re.sub(
        '[\u007E|\u00B4|\u0060|\u005E|\u00A8|\u0301|\u007E|\u005E|\xc2|\xb4|\xe9|\362]',
        '',
        texto_tratado,
    )

    # retorna o texto tratado
    return texto_limpo


def renomear(caminho, nome_atual, novo_nome):
    """Renomeia o nome de um arquivo no caminho informado."""
    # importa recursos do módulo Path
    from pathlib import Path

    # trata o caminho informado e o nome atual com o objeto Path
    nome_atual = Path(caminho) / nome_atual

    # trata o caminho informado e o nome novo com o objeto Path
    novo_nome = Path(caminho) / novo_nome

    # altera o nome atual para o nome novo
    novo_nome = nome_atual.rename(novo_nome)

    # retorna o caminho com o nome novo
    return str(novo_nome)


def retornar_arquivos_em_pasta(caminho, filtro='**/*'):
    """Retorna os arquivos existentes em um diretório se seus
    respectivos sub-diretórios segundo o caminho informado."""
    # importa recursos do módulo Path
    from pathlib import Path

    # coleta de forma recursiva o conteúdo
    #   contido no caminho informado caso existir
    lista_arquivos = list(Path(caminho).glob(filtro))

    # instancia uma lista vazia
    lista_arquivos_str = []

    # para cada arquivo na lista de arquivos
    for arquivo in lista_arquivos:
        # coleta e salva o arquivo em string
        lista_arquivos_str.append(str(arquivo))

    # retorna uma lista dos arquivos coletados
    return lista_arquivos_str


def retornar_data_hora_atual(parametro):
    """Formata e retorna dados de data e/ou hora,
    conforme informado pelo parâmetro."""
    # importa recursos do módulo datetime
    import datetime

    # retorna dados de data e/ou hora conforme informado pelo parâmetro.
    return datetime.datetime.now().strftime(parametro)



import os
import setuptools


base_dir = os.path.dirname(__file__)

with open(os.path.join(base_dir, "README.md"), encoding='utf8') as f:
    long_description = f.read()

setuptools.setup(
    name="py-rpautom",
    version="0.0.12b0",
    author="aranseiki",
    author_email="techall@hotmail.com.br",
    description="Conjunto de utilitários para automação de processos.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/aranseiki/py-rpautom/",
    packages=setuptools.find_packages(),
    project_urls={
        "Py-RPAutom": "https://aranseiki.github.io/py-rpautom/",
    },
    classifiers=[
        "Framework :: Robot Framework :: Library",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Development Status :: 4 - Beta",
        "Topic :: Utilities",
    ],
    python_requires=">=3.9",
    install_requires=[
        'openpyxl',
        'psutil',
        'pywinauto',
        'pywin32',
        'PyMuPDF',
        'PyPDF2',
        'Pytesseract',
        'requests',
        'selenium',
        'urllib3',
        'xlrd',
    ]
)



"""Módulo para automação web."""
from py_rpautom import desktop_utils as desktop_utils
from py_rpautom import python_utils as python_utils


__all__ = [
    'requisitar_url',
    'baixar_arquivo',
    'iniciar_navegador',
    'autenticar_navegador',
    'abrir_pagina',
    'abrir_pagina',
    'abrir_janela',
    'trocar_para',
    'coletar_id_janela',
    'coletar_todas_ids_janelas',
    'esperar_pagina_carregar',
    'voltar_pagina',
    'centralizar_elemento',
    'executar_script',
    'retornar_codigo_fonte',
    'aguardar_elemento',
    'procurar_muitos_elementos',
    'procurar_elemento',
    'selecionar_elemento',
    'contar_elementos',
    'extrair_texto',
    'coletar_atributo',
    'alterar_atributo',
    'clicar_elemento',
    'escrever_em_elemento',
    'limpar_campo',
    'performar',
    'print_para_pdf',
    'fechar_janela',
    'fechar_janelas_menos_essa',
    'encerrar_navegador',
]

def _coletar_tamanho(caminho):
    import os

    caminho_interno = python_utils.coletar_caminho_absoluto(caminho)

    return os.path.getsize(caminho_interno)


def _coletar_versao_navegador(caminho_arquivo):
    from win32api import GetFileVersionInfo, HIWORD, LOWORD

    caminho_interno = python_utils.coletar_caminho_absoluto(caminho_arquivo)
    informacao_arquivo = GetFileVersionInfo(caminho_interno, '\\')

    versao_major = informacao_arquivo['FileVersionMS']
    versao_minor = informacao_arquivo['FileVersionLS']

    versao_major_principal = str(HIWORD(versao_major))
    versao_major_secundario = str(LOWORD(versao_major))
    versao_minor_principal = str(HIWORD(versao_minor))
    versao_minor_secundario = str(LOWORD(versao_minor))

    versao = str('.').join(
        (
            versao_major_principal,
            versao_major_secundario,
            versao_minor_principal,
            versao_minor_secundario,
        )
    )

    return versao


def _coletar_versao_webdriver(executavel_webdriver):
    import subprocess

    execucao_webdriver = subprocess.Popen(
        [executavel_webdriver, '-V'], stdout=subprocess.PIPE
    )

    versao_webdriver = str(execucao_webdriver.stdout.read())
    versao_webdriver = versao_webdriver.partition(' (')[0]
    versao_webdriver = versao_webdriver.rpartition(' ')[-1]

    return versao_webdriver


def _preparar_ambiente(nome_navegador):
    # importa recursos dos módulos necessários
    import sys
    from collections import namedtuple

    caminho_auto_caminho = python_utils.coletar_arvore_caminho(__file__)
    pasta_projeto = python_utils.coletar_arvore_caminho(caminho_auto_caminho)
    caminho_webdriver_raiz = 'webdrivers'

    webdriver = namedtuple(
        'webdriver',
        [
            'nome',
            'versao',
            'nome_arquivo',
            'caminho',
            'tamanho',
            'plataforma',
        ],
    )

    webdriver.nome = ''
    webdriver.versao = ''
    webdriver.nome_arquivo = ''
    webdriver.caminho = ''
    webdriver.tamanho = ''
    webdriver.plataforma = ''

    if sys.platform == 'win32':
        webdriver.plataforma = 'win32'

    if nome_navegador.upper().__contains__('CHROME'):
        webdriver.nome_arquivo = 'chromedriver.exe'
        webdriver.nome = 'chromedriver'

        caminho_navegador = (
            'C:/Program Files (x86)/Google/Chrome/Application/chrome.exe'
        )
        url_base = 'https://chromedriver.storage.googleapis.com'

        lista_webdrivers = _coleta_lista_chromedrivers(url_base)

        tag_especificacao_abertura = '<Key>'
        tag_especificacao_encerramento = '</Key>'

        tag_tamanho_abertura = '<Size>'
        tag_tamanho_encerramento = '</Size>'
    elif nome_navegador.upper().__contains__('EDGE'):
        webdriver.nome_arquivo = 'msedgedriver.exe'
        webdriver.nome = 'edgedriver'

        caminho_navegador = (
            'C:/Program Files (x86)/Microsoft/Edge/Application/msedge.exe'
        )
        url_base = 'https://msedgedriver.azureedge.net/'

        lista_webdrivers = _coleta_lista_edgedrivers(url_base)

        tag_especificacao_abertura = '<Name>'
        tag_especificacao_encerramento = '</Name>'

        tag_tamanho_abertura = '<Content-Length>'
        tag_tamanho_encerramento = '</Content-Length>'
    elif nome_navegador.upper().__contains__('FIREFOX'):
        webdriver.nome_arquivo = 'geckodriver.exe'
        webdriver.nome = 'geckodriver'
        caminho_navegador = 'C:/Program Files/Mozilla Firefox/firefox.exe'
        url_base = 'https://github.com/mdn/content/blob/main/files/en-us/glossary/gecko/index.md?plain=1'
    else:
        raise SystemError(
            f' {nome_navegador} não disponível. Escolha uma dessas opções: Chrome, Edge, Firefox.'
        )

    webdriver.caminho = str('/').join(
        (
            pasta_projeto,
            caminho_webdriver_raiz,
            webdriver.nome,
        )
    )

    # caso o caminho existir
    if not python_utils.caminho_existente(webdriver.caminho):
        # cria a pasta informada, caso necessário
        #  cria a hierarquia anterior à última pasta
        python_utils.criar_pasta(webdriver.caminho)

    versao_navegador = _coletar_versao_navegador(caminho_navegador)

    lista_versoes = set()

    versao_sem_minor = versao_navegador[:-3]

    for grupo in lista_webdrivers.conteudo:
        especificacao = grupo[1]

        if (
            not especificacao.__contains__('.zip')
            or especificacao.__contains__('icons')
            or especificacao.__contains__('index.html')
            or especificacao.__contains__('icons/folder.gif')
            or especificacao.__contains__('LICENSE')
            or especificacao.__contains__('credits.html')
            or especificacao.__contains__('LATEST_RELEASE')
        ):
            continue

        especificacao = especificacao.removeprefix(tag_especificacao_abertura)
        especificacao = especificacao.removesuffix(
            tag_especificacao_encerramento
        )
        especificacao = especificacao.split('/')

        versao = especificacao[0]

        arquivo = especificacao[1]

        tamanho = grupo[6]
        tamanho = tamanho.removeprefix(tag_tamanho_abertura)
        tamanho = tamanho.removesuffix(tag_tamanho_encerramento)

        if versao.__contains__(versao_sem_minor):
            lista_versoes.add((versao, arquivo, tamanho))

    versoes_compativeis = []
    for versao in lista_versoes:
        versoes_compativeis.append(versao)

    versoes_compativeis.sort(reverse=True)

    for versao in versoes_compativeis:
        if (versao[0].__contains__(versao_navegador)) and (
            versao[1].__contains__(webdriver.plataforma)
        ):
            webdriver.versao = versao[0]
            arquivo_zip = versao[1]
            webdriver.tamanho = versao[2]
            break

    if webdriver.versao == '':
        for versao in versoes_compativeis:
            if versao[1].__contains__(webdriver.plataforma):
                break
        webdriver.versao = versoes_compativeis[0][0]
        arquivo_zip = versoes_compativeis[0][1]
        webdriver.tamanho = versoes_compativeis[0][2]

    url = str('/').join((url_base, webdriver.versao, arquivo_zip))

    caminho_arquivo_zip = str('/').join((webdriver.caminho, arquivo_zip))

    caminho_webdriver_pessoal = str('/').join(
        (
            webdriver.caminho,
            webdriver.versao,
        )
    )

    webdriver.caminho = python_utils.coletar_caminho_absoluto(
        caminho_webdriver_pessoal
    )

    # caso o caminho existir
    if not python_utils.caminho_existente(webdriver.caminho):
        # cria a pasta informada, caso necessário
        #  cria a hierarquia anterior à última pasta
        python_utils.criar_pasta(webdriver.caminho)

    caminho_arquivo_zip = python_utils.coletar_caminho_absoluto(caminho_arquivo_zip)

    return url, caminho_arquivo_zip, webdriver


def _coleta_lista_chromedrivers(url):
    from collections import namedtuple

    status = 0
    contagem = 0
    header_chrome = {
        'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.9',
        'Accept-Encoding': 'gzip, deflate, br',
        'Cache-Control': 'max-age=0',
        'Sec-Fetch-Dest': 'document',
        'Sec-Fetch-Mode': 'navigate',
        'Sec-Fetch-Site': 'none',
        'Sec-Fetch-User': '?1',
        'Upgrade-Insecure-Requests': '1',
    }
    while not status == 200 and contagem < 60:
        resposta = requisitar_url(url, stream=True, header_arg=header_chrome)
        status = resposta.status_code
        contagem = contagem + 1

    if not resposta.status_code in range(200, 300):
        raise SystemError(
            f'Falha ao acessar a url {url}. Revise os dados e tente novamente.'
        )

    resposta_str = str(resposta.content)
    resposta_str_dividido = resposta_str.replace('><', '>\n<')
    resposta_str_dividido = resposta_str_dividido.split('\n')

    lista_webdrivers = namedtuple(
        'lista_webdrivers', ['cabecalho', 'conteudo', 'rodape']
    )

    lista_webdrivers.cabecalho = []
    lista_webdrivers.conteudo = []
    lista_webdrivers.rodape = []

    versao_xml = resposta_str_dividido[0]
    abertura_xmlns = resposta_str_dividido[1]
    nome = resposta_str_dividido[2]
    abertura_prefix = resposta_str_dividido[3]
    encerramento_prefix = resposta_str_dividido[4]
    abertura_maker = resposta_str_dividido[5]
    encerramento_maker = resposta_str_dividido[6]
    validacao_truncado = resposta_str_dividido[7]

    lista_webdrivers.cabecalho.append(
        [
            versao_xml,
            abertura_xmlns,
            nome,
            abertura_prefix,
            encerramento_prefix,
            abertura_maker,
            encerramento_maker,
            validacao_truncado,
        ]
    )

    for item in (
        versao_xml,
        abertura_xmlns,
        nome,
        abertura_prefix,
        encerramento_prefix,
        abertura_maker,
        encerramento_maker,
        validacao_truncado,
    ):
        resposta_str_dividido.remove(item)

    encerramento_xmlns = resposta_str_dividido[-1]

    lista_webdrivers.rodape.append([encerramento_xmlns])

    resposta_str_dividido.remove(encerramento_xmlns)

    total = int(len(resposta_str_dividido) / 8)
    contagem = 0
    while contagem < total:
        abertura_contents = resposta_str_dividido[0]
        key = resposta_str_dividido[1]
        generation = resposta_str_dividido[2]
        meta_generation = resposta_str_dividido[3]
        last_modified = resposta_str_dividido[4]
        etag = resposta_str_dividido[5]
        size = resposta_str_dividido[6]
        fechamento_contents = resposta_str_dividido[7]

        for item in (
            abertura_contents,
            key,
            generation,
            meta_generation,
            last_modified,
            etag,
            size,
            fechamento_contents,
        ):
            resposta_str_dividido.remove(item)

        lista_webdrivers.conteudo.append(
            [
                abertura_contents,
                key,
                generation,
                meta_generation,
                last_modified,
                etag,
                size,
                fechamento_contents,
            ]
        )

        contagem = contagem + 1

    return lista_webdrivers


def _coleta_lista_edgedrivers(url):
    from collections import namedtuple

    status = 0
    contagem = 0
    header_edge = {
        'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.9',
        'Accept-Encoding': 'gzip, deflate, br',
        'Cache-Control': 'max-age=0',
        'Sec-Fetch-Dest': 'document',
        'Sec-Fetch-Mode': 'navigate',
        'Sec-Fetch-Site': 'none',
        'Sec-Fetch-User': '?1',
        'Upgrade-Insecure-Requests': '1',
    }
    while not status == 200 and contagem < 60:
        resposta = requisitar_url(
            url, stream=True, autenticacao=['', ''], header_arg=header_edge
        )
        status = resposta.status_code

        if status in range(200, 300):
            break
        else:
            resposta = requisitar_url(
                url,
                stream=True,
                header_arg=header_edge,
            )
            status = resposta.status_code

        contagem = contagem + 1

    if not resposta.status_code in range(200, 300):
        raise SystemError(
            f'Falha ao acessar a url {url}. Revise os dados e tente novamente.'
        )

    resposta_str = str(resposta.content)
    resposta_str_dividido = resposta_str.replace('><', '>\n<')
    resposta_str_dividido = resposta_str_dividido.split('\n')

    lista_webdrivers = namedtuple(
        'lista_webdrivers', ['cabecalho', 'conteudo', 'rodape']
    )

    lista_webdrivers.cabecalho = []
    lista_webdrivers.conteudo = []
    lista_webdrivers.rodape = []

    versao_xml = resposta_str_dividido[0]
    abertura_enumeration_results = resposta_str_dividido[1]
    abertura_blobs = resposta_str_dividido[2]

    lista_webdrivers.cabecalho.append(
        [
            versao_xml,
            abertura_enumeration_results,
            abertura_blobs,
        ]
    )

    for item in (
        versao_xml,
        abertura_enumeration_results,
        abertura_blobs,
    ):
        resposta_str_dividido.remove(item)

    encerramento_blobs = resposta_str_dividido[-3]
    encerramento_next_marker = resposta_str_dividido[-2]
    encerramento_enumeration_results = resposta_str_dividido[-1]

    lista_webdrivers.rodape.append(
        [
            encerramento_blobs,
            encerramento_next_marker,
            encerramento_enumeration_results,
        ]
    )

    for item in (
        encerramento_blobs,
        encerramento_next_marker,
        encerramento_enumeration_results,
    ):
        resposta_str_dividido.remove(item)

    total = int(len(resposta_str_dividido) / 16)
    contagem = 0

    while contagem < total:
        abertura_blob = resposta_str_dividido[0]
        name = resposta_str_dividido[1]
        url = resposta_str_dividido[2]
        properties = resposta_str_dividido[3]
        last_modified = resposta_str_dividido[4]
        e_tag = resposta_str_dividido[5]
        content_length = resposta_str_dividido[6]
        content_type = resposta_str_dividido[7]
        content_encoding = resposta_str_dividido[8]
        content_language = resposta_str_dividido[9]
        content_mds = resposta_str_dividido[10]
        cache_control = resposta_str_dividido[11]
        blob_type = resposta_str_dividido[12]
        lease_status = resposta_str_dividido[13]
        encerramento_properties = resposta_str_dividido[14]
        encerramento_blob = resposta_str_dividido[15]

        for item in (
            abertura_blob,
            name,
            url,
            properties,
            last_modified,
            e_tag,
            content_length,
            content_type,
            content_encoding,
            content_language,
            content_mds,
            cache_control,
            blob_type,
            lease_status,
            encerramento_properties,
            encerramento_blob,
        ):
            resposta_str_dividido.remove(item)

        lista_webdrivers.conteudo.append(
            [
                abertura_blob,
                name,
                url,
                properties,
                last_modified,
                e_tag,
                content_length,
                content_type,
                content_encoding,
                content_language,
                content_mds,
                cache_control,
                blob_type,
                lease_status,
                encerramento_properties,
                encerramento_blob,
            ]
        )

        contagem = contagem + 1

    return lista_webdrivers


def _escolher_tipo_elemento(tipo_elemento):
    """Escolhe um tipo de elemento 'locator'."""
    from selenium.webdriver.common.by import By

    if tipo_elemento.upper() == 'CLASS_NAME':
        tipo_elemento = By.CLASS_NAME
    elif tipo_elemento.upper() == 'CSS_SELECTOR':
        tipo_elemento = By.CSS_SELECTOR
    elif tipo_elemento.upper() == 'ID':
        tipo_elemento = By.ID
    elif tipo_elemento.upper() == 'LINK_TEXT':
        tipo_elemento = By.LINK_TEXT
    elif tipo_elemento.upper() == 'NAME':
        tipo_elemento = By.NAME
    elif tipo_elemento.upper() == 'PARTIAL_LINK_TEXT':
        tipo_elemento = By.PARTIAL_LINK_TEXT
    elif tipo_elemento.upper() == 'TAG_NAME':
        tipo_elemento = By.TAG_NAME
    elif tipo_elemento.upper() == 'XPATH':
        tipo_elemento = By.XPATH
    return tipo_elemento


def _escolher_comportamento_esperado(comportamento_esperado: str):
    """Escolhe um tipo de comportamento manipulado pelo Selenium."""
    from selenium.webdriver.support import expected_conditions as EC

    if comportamento_esperado.upper() == 'ALERT_IS_PRESENT':
        comportamento_esperado = EC.alert_is_present
    elif comportamento_esperado.upper() == 'ALL_OF':
        comportamento_esperado = EC.all_of
    elif comportamento_esperado.upper() == 'ANY_OF':
        comportamento_esperado = EC.any_of
    elif comportamento_esperado.upper() == 'ELEMENT_ATTRIBUTE_TO_INCLUDE':
        comportamento_esperado = EC.element_attribute_to_include
    elif (
        comportamento_esperado.upper()
        == 'ELEMENT_LOCATED_SELECTION_STATE_TO_BE'
    ):
        comportamento_esperado = EC.element_located_selection_state_to_be
    elif comportamento_esperado.upper() == 'ELEMENT_LOCATED_TO_BE_SELECTED':
        comportamento_esperado = EC.element_located_to_be_selected
    elif comportamento_esperado.upper() == 'ELEMENT_SELECTION_STATE_TO_BE':
        comportamento_esperado = EC.element_selection_state_to_be
    elif comportamento_esperado.upper() == 'ELEMENT_TO_BE_CLICKABLE':
        comportamento_esperado = EC.element_to_be_clickable
    elif comportamento_esperado.upper() == 'ELEMENT_TO_BE_SELECTED':
        comportamento_esperado = EC.element_to_be_selected
    elif (
        comportamento_esperado.upper()
        == 'FRAME_TO_BE_AVAILABLE_AND_SWITCH_TO_IT'
    ):
        comportamento_esperado = EC.frame_to_be_available_and_switch_to_it
    elif comportamento_esperado.upper() == 'INVISIBILITY_OF_ELEMENT':
        comportamento_esperado = EC.invisibility_of_element
    elif comportamento_esperado.upper() == 'INVISIBILITY_OF_ELEMENT_LOCATED':
        comportamento_esperado = EC.invisibility_of_element_located
    elif comportamento_esperado.upper() == 'NEW_WINDOW_IS_OPENED':
        comportamento_esperado = EC.new_window_is_opened
    elif comportamento_esperado.upper() == 'NONE_OF':
        comportamento_esperado = EC.none_of
    elif comportamento_esperado.upper() == 'NUMBER_OF_WINDOWS_TO_BE':
        comportamento_esperado = EC.number_of_windows_to_be
    elif comportamento_esperado.upper() == 'PRESENCE_OF_ALL_ELEMENTS_LOCATED':
        comportamento_esperado = EC.presence_of_all_elements_located
    elif comportamento_esperado.upper() == 'PRESENCE_OF_ELEMENT_LOCATED':
        comportamento_esperado = EC.presence_of_element_located
    elif comportamento_esperado.upper() == 'STALENESS_OF':
        comportamento_esperado = EC.staleness_of
    elif comportamento_esperado.upper() == 'TEXT_TO_BE_PRESENT_IN_ELEMENT':
        comportamento_esperado = EC.text_to_be_present_in_element
    elif (
        comportamento_esperado.upper()
        == 'TEXT_TO_BE_PRESENT_IN_ELEMENT_ATTRIBUTE'
    ):
        comportamento_esperado = EC.text_to_be_present_in_element_attribute
    elif (
        comportamento_esperado.upper() == 'TEXT_TO_BE_PRESENT_IN_ELEMENT_VALUE'
    ):
        comportamento_esperado = EC.text_to_be_present_in_element_value
    elif comportamento_esperado.upper() == 'TITLE_CONTAINS':
        comportamento_esperado = EC.title_contains
    elif comportamento_esperado.upper() == 'TITLE_IS':
        comportamento_esperado = EC.title_is
    elif comportamento_esperado.upper() == 'URL_CHANGES':
        comportamento_esperado = EC.url_changes
    elif comportamento_esperado.upper() == 'URL_CONTAINS':
        comportamento_esperado = EC.url_contains
    elif comportamento_esperado.upper() == 'URL_MATCHES':
        comportamento_esperado = EC.url_matches
    elif comportamento_esperado.upper() == 'URL_TO_BE':
        comportamento_esperado = EC.url_to_be
    elif comportamento_esperado.upper() == 'VISIBILITY_OF':
        comportamento_esperado = EC.visibility_of
    elif (
        comportamento_esperado.upper() == 'VISIBILITY_OF_ALL_ELEMENTS_LOCATED'
    ):
        comportamento_esperado = EC.visibility_of_all_elements_located
    elif (
        comportamento_esperado.upper() == 'VISIBILITY_OF_ANY_ELEMENTS_LOCATED'
    ):
        comportamento_esperado = EC.visibility_of_any_elements_located
    elif comportamento_esperado.upper() == 'VISIBILITY_OF_ELEMENT_LOCATED':
        comportamento_esperado = EC.visibility_of_element_located
    return comportamento_esperado


def _procurar_elemento(seletor, tipo_elemento='CSS_SELECTOR'):
    """Procura um elemento presente que corresponda ao informado."""
    tipo_elemento = _escolher_tipo_elemento(tipo_elemento)
    webelemento = _navegador.find_element(tipo_elemento, seletor)
    centralizar_elemento(seletor, tipo_elemento)

    return webelemento


def _procurar_muitos_elementos(seletor, tipo_elemento='CSS_SELECTOR'):
    """Procura todos os elementos presentes que correspondam ao informado."""
    # instancia uma lista vazia
    lista_webelementos_str = []

    tipo_elemento = _escolher_tipo_elemento(tipo_elemento)
    lista_webelementos = _navegador.find_elements(tipo_elemento, seletor)
    centralizar_elemento(seletor, tipo_elemento)

    # retorna os valores coletados ou uma lista vazia
    return lista_webelementos


def requisitar_url(
    url: str,
    stream=True,
    autenticacao: None or list = None,
    header_arg: str = None,
):
    """Faz uma requisição http, retornando a resposta
    dessa requisição no padrão http/https."""
    import os
    from requests import get
    from requests.auth import HTTPBasicAuth

    verificacao_ssl = (
        os.environ['WDM_SSL_VERIFY']
    ).lower() in ['1', 1, 'true', True,]

    if autenticacao is not None:
        usuario, senha = autenticacao
        autenticacao = HTTPBasicAuth(usuario, senha)

    resposta = get(
        url,
        stream=stream,
        verify=verificacao_ssl,
        auth=autenticacao,
        headers=header_arg,
    )

    return resposta


def baixar_arquivo(
    url,
    caminho_destino,
    stream=True,
    autenticacao: None or list = None,
    header_arg=None,
) -> bool:
    """Baixa um arquivo mediante uma url do arquivo e um
    caminho de destino já com o nome do arquivo a ser gravado."""
    from shutil import copyfileobj

    caminho_interno_absoluto = python_utils.coletar_caminho_absoluto(
        caminho_destino
    )

    resposta = requisitar_url(
        url,
        stream=stream,
        autenticacao=autenticacao,
        header_arg=header_arg,
    )
    with open(caminho_interno_absoluto, 'wb') as code:
        copyfileobj(resposta.raw, code)

    return True


def validar_porta(ip, porta, tempo_limite=1):
    import socket
    conexao = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    conexao.settimeout(tempo_limite)
    retorno_validacao = conexao.connect_ex((ip, porta))

    if retorno_validacao == 0:
        return True

    return False


def iniciar_navegador(
    url: str,
    nome_navegador: str,
    options: tuple = (''),
    extensoes: tuple = (''),
    experimentos: tuple = (''),
    capacidades: tuple = (''),
    executavel: str = 'padrao',
    porta_webdriver: int = None,
    baixar_webdriver_previamente: bool = True,
):
    """Inicia uma instância automatizada de um navegador."""
    import urllib3
    from selenium import webdriver
    from random import randint


    urllib3.disable_warnings()

    global _navegador
    com_extras = False


    if porta_webdriver is None:
        porta_webdriver = 0
        validacao_porta = None
        while validacao_porta is not False:
            limite_porta_minima = 49152
            limite_porta_maxima = 65535
            porta_webdriver = randint(limite_porta_minima, limite_porta_maxima)
            validacao_porta = validar_porta('localhost', porta_webdriver)
    else:
        if isinstance(porta_webdriver, int) is False:
            raise ValueError(
                'Parâmetro ``porta_webdriver`` precisa ser número e do tipo inteiro.'
            )

    def retorna_service(executavel_webdriver, nome_navegador, porta_webdriver):
        if nome_navegador.upper().__contains__('CHROME'):
            from selenium.webdriver.chrome.service import Service
        if nome_navegador.upper().__contains__('EDGE'):
            from selenium.webdriver.edge.service import Service
        if nome_navegador.upper().__contains__('FIREFOX'):
            from selenium.webdriver.firefox.service import Service

        executavel = python_utils.coletar_caminho_absoluto(executavel_webdriver)

        service = Service(executable_path=executavel, port=porta_webdriver)
        return service

    def adicionar_extras(
        options_webdriver,
        argumento,
        extensao,
        argumento_experimental,
        capacidade,
    ):
        if len(argumento) > 0:
            for item in argumento:
                options_webdriver.add_argument(item)

        if len(extensao) > 0:
            for item in extensao:
                options_webdriver.add_extension(item)

        if len(argumento_experimental) > 0:
            for item in argumento_experimental:
                options_webdriver.add_experimental_option(*item)

        if len(capacidade) > 0:
            for item in capacidade:
                options_webdriver.set_capability(item[0], item[1])
            options_webdriver.to_capabilities()

        return options_webdriver

    if baixar_webdriver_previamente is True:
        (
            url_webdriver,
            caminho_arquivo_zip,
            informacao_webdriver,
        ) = _preparar_ambiente(nome_navegador)

        executavel_webdriver = str('/').join(
            (
                informacao_webdriver.caminho,
                informacao_webdriver.nome_arquivo,
            )
        )

        executavel_webdriver = python_utils.coletar_caminho_absoluto(
            executavel_webdriver
        )

        if not python_utils.caminho_existente(executavel_webdriver):
            validacao_baixar_arquivo = baixar_arquivo(
                url_webdriver,
                caminho_arquivo_zip,
            )

            validacao_caminho_arquivo_zip = 0
            while not (
                validacao_caminho_arquivo_zip == informacao_webdriver.tamanho
            ):
                validacao_caminho_arquivo_zip = str(
                    _coletar_tamanho(caminho_arquivo_zip)
                )

            caminho_descompactar = python_utils.coletar_caminho_absoluto(
                informacao_webdriver.caminho
            )

            python_utils.descompactar(
                arquivo=caminho_arquivo_zip,
                caminho_destino=caminho_descompactar,
            )
        else:
            executavel_webdriver = str(executavel_webdriver)
    else:
        if executavel == 'padrao':
            raise SystemError('Informe o executável do webdriver.')
        else:
            if not executavel.__contains__('.exe'):
                raise SystemError('Informe o executável do webdriver.')
            else:
                executavel_webdriver = python_utils.coletar_caminho_absoluto(
                    executavel
                )

    if nome_navegador.upper().__contains__('CHROME'):
        options_webdriver = webdriver.ChromeOptions()
        instancia_webdriver = webdriver.Chrome
    elif nome_navegador.upper().__contains__('EDGE'):
        options_webdriver = webdriver.EdgeOptions()
        instancia_webdriver = webdriver.Edge
    elif nome_navegador.upper().__contains__('FIREFOX'):
        options_webdriver = webdriver.FirefoxOptions()
        instancia_webdriver = webdriver.Firefox
    else:
        raise SystemError(
            f' {nome_navegador} não disponível. Escolha uma dessas opções: Chrome, Edge, Firefox.'
        )

    options_webdriver = adicionar_extras(
        options_webdriver=options_webdriver,
        argumento=options,
        extensao=extensoes,
        argumento_experimental=experimentos,
        capacidade=capacidades,
    )

    _navegador = instancia_webdriver(
        service=retorna_service(
            executavel_webdriver,
            nome_navegador,
            porta_webdriver,
        ),
        options=options_webdriver,
    )

    abrir_pagina(url)

    return True


def autenticar_navegador(
    nome_navegador: str,
    titulo_janela: str,
    usuario: str,
    senha: str,
) -> bool:
    from time import sleep

    lista_processos_navegadores = (
        'EDGE',
        'CHROME',
        'FIREFOX',
    )

    if not nome_navegador.upper() in lista_processos_navegadores:
        raise NameError(
            'Escolha um desses nomes de navegador: Edge, Chrome, Firefox.'
        )

    nome_janela_navegador = ''
    pid_janela_credenciais = python_utils.coletar_pid(nome_navegador)
    lista_relacao_janelas = []
    pid_janela_navegador = 0
    nome_janela_navegador = ''
    for indice_lista_janela in range(0, len(pid_janela_credenciais)):
        try:
            pid = pid_janela_credenciais[indice_lista_janela]['pid']
            desktop_utils.conectar_app(
                pid, estilo_aplicacao='uia', tempo_espera=1
            )
            lista_relacao_janelas.append(
                [pid, desktop_utils.retornar_janelas_disponiveis(pid)]
            )
        except:
            pass

    for janela in range(0, len(lista_relacao_janelas)):
        if not lista_relacao_janelas[janela][1] == []:
            pid_janela_navegador = lista_relacao_janelas[janela][0]
            nome_janela_navegador = lista_relacao_janelas[janela][1][0]
            if nome_janela_navegador.__contains__(titulo_janela):
                break

    desktop_utils.conectar_app(pid_janela_navegador)
    desktop_utils.ativar_foco(nome_janela_navegador)

    desktop_utils.simular_digitacao(usuario)
    desktop_utils.simular_digitacao('{TAB}')
    desktop_utils.simular_digitacao(senha)
    desktop_utils.simular_digitacao('{TAB}')
    desktop_utils.simular_digitacao('{ENTER}')

    return True


def abrir_pagina(url: str):
    """Abre uma página web mediante a URL informada."""
    global _navegador

    _navegador.get(url)
    esperar_pagina_carregar()


def abrir_janela(url: str = None):
    """Abre uma nova janela/aba do navegador automatizado."""
    _navegador.window_handles
    _navegador.execute_script(f'window.open("{url}")')


def trocar_para(id, tipo):
    """Troca de contexto da automação web mediante o tipo e o id informados."""

    try:
        resultado = True
        if tipo.upper() == 'FRAME':
            _navegador.switch_to.frame(id)
        elif tipo.upper() == 'PARENT_FRAME':
            _navegador.switch_to.parent_frame(id)
        elif tipo.upper() == 'NEW_WINDOW':
            _navegador.switch_to.new_window(id)
        elif tipo.upper() == 'WINDOW':
            _navegador.switch_to.window(id)
        elif tipo.upper() == 'ALERT':
            if id.upper() == 'TEXT':
                resultado = _navegador.switch_to.alert.text
            elif id.upper() == 'DISMISS':
                _navegador.switch_to.alert.dismiss()
            elif id.upper() == 'ACCEPT':
                _navegador.switch_to.alert.accept()
            elif id.upper().__contains__('SEND_KEYS'):
                metodo, valor = id
                _navegador.switch_to.alert.send_keys(valor)
            else:
                _navegador.switch_to.alert.accept()
        elif tipo.upper() == 'ACTIVE_ELEMENT':
            _navegador.switch_to.active_element(id)
        elif tipo.upper() == 'DEFAULT_CONTENT':
            _navegador.switch_to.default_content()
        try:
            esperar_pagina_carregar()
        except:
            ...
        return resultado
    except:
        return False


def coletar_id_janela():
    """Coleta um ID de uma janela/aba."""
    id_janela = _navegador.current_window_handle

    return id_janela


def coletar_todas_ids_janelas():
    """Coleta uma lista de todos os ID's de
    janelas/abas do navegador automatizado."""
    ids_janelas = _navegador.window_handles

    return ids_janelas


def esperar_pagina_carregar():
    """Espera o carregamento total da página automatizada acontecer."""

    estado_pronto = False
    while estado_pronto is False:
        state = _navegador.execute_script('return window.document.readyState')

        if state == 'complete':
            estado_pronto = True


def voltar_pagina():
    """Volta o contexto de histórico do navegador automatizado."""
    _navegador.back()

    esperar_pagina_carregar()


def centralizar_elemento(seletor, tipo_elemento='CSS_SELECTOR'):
    """Centraliza um elemento informado na tela."""
    seletor = seletor.replace('"', "'")

    if tipo_elemento.upper() == 'CSS_SELECTOR':
        _navegador.execute_script(
            'document.querySelector("'
            + seletor
            + "\").scrollIntoView({block:  'center'})"
        )
    elif tipo_elemento.upper() == 'XPATH':
        _navegador.execute_script(
            'elemento = document.evaluate("'
            + seletor
            + "\", document, null, XPathResult.FIRST_ORDERED_NODE_TYPE, null)\
            .singleNodeValue; elemento.scrollIntoView({block: 'center'});"
        )


def executar_script(script, args=''):
    """Executa um script Javascript na página automatizada."""
    try:
        if not args == '':
            _navegador.execute_script(script, args)
        else:
            _navegador.execute_script(script)

        return True
    except:
        return False


def retornar_codigo_fonte():
    """Coleta e retorna o código HTML da página automatizada."""
    codigo_fonte = _navegador.page_source

    return codigo_fonte


def aguardar_elemento(
    identificador: str or int,
    tipo_elemento: str = 'CSS_SELECTOR',
    valor: str or tuple or bool = '',
    comportamento_esperado: str = 'VISIBILITY_OF_ELEMENT_LOCATED',
    tempo: int = 30,
):
    """Aguarda o elemento informado estar visível na tela."""
    from selenium.webdriver.support.ui import WebDriverWait

    _lista_ec_sem_parametro = ['ALERT_IS_PRESENT']
    _lista_ec_com_locator = [
        'ELEMENT_LOCATED_TO_BE_SELECTED',
        'FRAME_TO_BE_AVAILABLE_AND_SWITCH_TO_IT',
        'INVISIBILITY_OF_ELEMENT_LOCATED',
        'PRESENCE_OF_ALL_ELEMENTS_LOCATED',
        'PRESENCE_OF_ELEMENT_LOCATED',
        'VISIBILITY_OF_ALL_ELEMENTS_LOCATED',
        'VISIBILITY_OF_ANY_ELEMENTS_LOCATED',
        'VISIBILITY_OF_ELEMENT_LOCATED',
        'ELEMENT_TO_BE_CLICKABLE',
    ]
    _lista_ec_com_locator_texto = [
        'TEXT_TO_BE_PRESENT_IN_ELEMENT',
        'TEXT_TO_BE_PRESENT_IN_ELEMENT_VALUE',
    ]
    _lista_ec_com_locator_boleano = ['ELEMENT_LOCATED_SELECTION_STATE_TO_BE']
    _lista_ec_com_element_boleano = ['ELEMENT_SELECTION_STATE_TO_BE']
    _lista_ec_com_locator_atributo = ['ELEMENT_ATTRIBUTE_TO_INCLUDE']
    _lista_ec_com_locator_atributo_texto = [
        'TEXT_TO_BE_PRESENT_IN_ELEMENT_ATTRIBUTE'
    ]
    _lista_ec_com_titulo = ['TITLE_CONTAINS', 'TITLE_IS']
    _lista_ec_com_url = ['URL_CHANGES', 'URL_CONTAINS', 'URL_TO_BE']
    _lista_ec_com_pattern = ['URL_MATCHES']
    _lista_ec_com_element = [
        'ELEMENT_TO_BE_SELECTED',
        'INVISIBILITY_OF_ELEMENT',
        'STALENESS_OF',
        'VISIBILITY_OF',
    ]
    _lista_ec_com_ec = ['ALL_OF', 'ANY_OF', 'NONE_OF']
    _lista_ec_com_handle = ['NEW_WINDOW_IS_OPENED']
    _lista_ec_com_int = ['NUMBER_OF_WINDOWS_TO_BE']

    try:
        wait = WebDriverWait(_navegador, tempo)

        tipo_elemento_escolhido = _escolher_tipo_elemento(tipo_elemento)
        tipo_comportamento_esperado = _escolher_comportamento_esperado(
            comportamento_esperado
        )
        complemento = False

        if (identificador == '') and (tipo_elemento == ''):
            if comportamento_esperado in _lista_ec_sem_parametro:
                wait.until(tipo_comportamento_esperado())
            elif comportamento_esperado in _lista_ec_com_handle:
                wait.until(
                    tipo_comportamento_esperado(_navegador.window_handles)
                )
        elif (
            (identificador == '')
            or (tipo_elemento == '')
            or (comportamento_esperado in _lista_ec_com_ec)
        ):
            raise
        else:
            if (comportamento_esperado in _lista_ec_com_locator) or (
                comportamento_esperado in _lista_ec_com_element
            ):
                wait.until(
                    tipo_comportamento_esperado(
                        (tipo_elemento_escolhido, identificador)
                    )
                )
                complemento = True
            elif (
                (comportamento_esperado in _lista_ec_com_locator_texto)
                or (comportamento_esperado in _lista_ec_com_locator_atributo)
                or (comportamento_esperado in _lista_ec_com_locator_boleano)
            ):
                wait.until(
                    tipo_comportamento_esperado(
                        (tipo_elemento_escolhido, identificador), valor
                    )
                )
                complemento = True
            elif comportamento_esperado in _lista_ec_com_element_boleano:
                wait.until(
                    tipo_comportamento_esperado(
                        _procurar_elemento(
                            tipo_elemento_escolhido, identificador
                        ),
                        valor,
                    )
                )
                complemento = True
            elif (
                comportamento_esperado in _lista_ec_com_locator_atributo_texto
            ):
                atributo = valor[0]
                texto = valor[1]
                wait.until(
                    tipo_comportamento_esperado(
                        (tipo_elemento_escolhido, identificador),
                        atributo,
                        texto,
                    )
                )
                complemento = True
            elif (
                (comportamento_esperado in _lista_ec_com_titulo)
                or (comportamento_esperado in _lista_ec_com_url)
                or (comportamento_esperado in _lista_ec_com_pattern)
                or (comportamento_esperado in _lista_ec_com_int)
            ):
                wait.until(tipo_comportamento_esperado(identificador))
            else:
                raise

        if complemento is True:
            centralizar_elemento(identificador, tipo_elemento)
            esperar_pagina_carregar()

        return True
    except:
        return False


def procurar_muitos_elementos(seletor, tipo_elemento='CSS_SELECTOR'):
    """Procura todos os elementos presentes que correspondam ao informado."""
    # instancia uma lista vazia
    lista_webelementos_str = []

    tipo_elemento = _escolher_tipo_elemento(tipo_elemento)
    lista_webelementos = _navegador.find_elements(tipo_elemento, seletor)
    centralizar_elemento(seletor, tipo_elemento)

    # para cada elemento na lista de webelementos
    for webelemento in lista_webelementos:
        # coleta e salva o texto do elemento
        lista_webelementos_str.append(webelemento.text)

    # retorna os valores coletados ou uma lista vazia
    return lista_webelementos_str


def procurar_elemento(seletor, tipo_elemento='CSS_SELECTOR'):
    """Procura um elemento presente que corresponda ao informado."""
    try:
        tipo_elemento = _escolher_tipo_elemento(tipo_elemento)
        _navegador.find_element(tipo_elemento, seletor)
        centralizar_elemento(seletor, tipo_elemento)

        return True
    except Exception:
        return False


def selecionar_elemento(
    seletor: str,
    valor: str,
    tipo_elemento: str = 'CSS_SELECTOR',
):
    """Seleciona em elemento de seleção um
    valor que corresponda ao informado."""
    from selenium.webdriver.support.ui import Select

    tipo_elemento = _escolher_tipo_elemento(tipo_elemento)
    aguardar_elemento(seletor, tipo_elemento)

    webelemento = _procurar_elemento(
        seletor,
        tipo_elemento,
    )

    Select(webelemento).select_by_visible_text(valor)

    return True


def contar_elementos(seletor, tipo_elemento='CSS_SELECTOR'):
    """Conta todos os elementos presentes que correspondam ao informado."""
    tipo_elemento = _escolher_tipo_elemento(tipo_elemento)
    centralizar_elemento(seletor, tipo_elemento)
    elementos = _procurar_muitos_elementos(seletor, tipo_elemento)

    return len(elementos)


def extrair_texto(seletor, tipo_elemento='CSS_SELECTOR'):
    """Extrai o texto de um elemento informado."""
    tipo_elemento = _escolher_tipo_elemento(tipo_elemento)
    elemento = _procurar_elemento(seletor, tipo_elemento)
    text_element = elemento.text
    centralizar_elemento(seletor, tipo_elemento)

    return text_element


def coletar_atributo(seletor, atributo, tipo_elemento='CSS_SELECTOR'):
    """Coleta o valor de um atributo solicitado do elemento informado."""
    tipo_elemento = _escolher_tipo_elemento(tipo_elemento)
    elemento = _procurar_elemento(seletor, tipo_elemento)
    centralizar_elemento(seletor, tipo_elemento)
    valor_atributo = elemento.get_attribute(atributo)

    return valor_atributo


def alterar_atributo(
    seletor,
    atributo,
    novo_valor,
    tipo_elemento='CSS_SELECTOR',
):
    """Coleta o valor de um atributo solicitado do elemento informado."""
    seletor = seletor.replace('"', "'")
    tipo_elemento_transformado = _escolher_tipo_elemento(tipo_elemento)
    elemento = _procurar_elemento(seletor, tipo_elemento_transformado)
    centralizar_elemento(seletor, tipo_elemento_transformado)

    if tipo_elemento.upper() == 'XPATH':
        _navegador.execute_script(
            f"""elemento_xpath = document.evaluate(\"{seletor}\",
            document, null, XPathResult.FIRST_ORDERED_NODE_TYPE,
            null).singleNodeValue;elemento.{atributo} = \"{novo_valor}\""""
        )
    elif tipo_elemento.upper() == 'CSS_SELECTOR':
        _navegador.execute_script(
            f"""elemento = document.querySelector(\"{seletor}\");
            elemento.{atributo} = \"{novo_valor}\""""
        )

    valor_atributo = elemento.get_attribute(atributo)

    return valor_atributo


def clicar_elemento(seletor, tipo_elemento='CSS_SELECTOR'):
    """Clica em um elemento informado."""
    tipo_elemento = _escolher_tipo_elemento(tipo_elemento)
    centralizar_elemento(seletor, tipo_elemento)

    elemento = _procurar_elemento(seletor, tipo_elemento)
    elemento.click()
    esperar_pagina_carregar()


def escrever_em_elemento(
    seletor, texto, tipo_elemento='CSS_SELECTOR', performar: bool = False
):
    """Digita dentro de um elemento informado."""
    from selenium.webdriver import ActionChains

    tipo_elemento = _escolher_tipo_elemento(tipo_elemento)

    try:
        centralizar_elemento(seletor, tipo_elemento)
    except:
        ...

    webelemento = _procurar_elemento(seletor, tipo_elemento=tipo_elemento)

    if performar is False:
        webelemento.send_keys(texto)
    else:
        action = ActionChains(_navegador)
        action.click(webelemento).perform()
        webelemento.clear()
        action.send_keys_to_element(webelemento, texto).perform()

    esperar_pagina_carregar()


def limpar_campo(
    seletor, tipo_elemento='CSS_SELECTOR', performar: bool = False
):
    """Digita dentro de um elemento informado."""
    from selenium.webdriver import ActionChains

    tipo_elemento = _escolher_tipo_elemento(tipo_elemento)
    centralizar_elemento(seletor, tipo_elemento)

    webelemento = _procurar_elemento(seletor, tipo_elemento=tipo_elemento)
    if performar is False:
        webelemento.clear()
    else:
        action = ActionChains(_navegador)
        action.click(webelemento).perform()
        webelemento.clear()
        action.send_keys_to_element(webelemento, '').perform()
        action.reset_actions()

    esperar_pagina_carregar()


def performar(acao, seletor, tipo_elemento='CSS_SELECTOR'):
    """Simula uma ação real de mouse."""
    from selenium.webdriver import ActionChains

    action = ActionChains(_navegador)
    webelemento = _procurar_elemento(seletor, tipo_elemento)

    if acao.upper() == 'CLICK':
        action.click(webelemento).perform()
    elif acao.upper() == 'DOUBLE_CLICK':
        action.double_click(webelemento).perform()
        action = ActionChains(_navegador)
    elif acao.upper() == 'MOVE_TO_ELEMENT':
        action.move_to_element(webelemento).perform()

    return True


def print_para_pdf(
    caminho_arquivo: str,
    escala: float = 1.0,
    paginacao: list[str] = None,
    fundo: bool = None,
    encolher_para_caber: bool = None,
    orientacao: int = None,
):
    """Realiza o print da página atual e salva em um arquivo informado."""
    # importa recursos do módulo base64
    import base64

    # importa recursos do módulo print_page_options
    from selenium.webdriver.common.print_page_options import PrintOptions

    try:
        # coleta o caminho completo do arquivo informado
        caminho_arquivo_absoluto = python_utils.coletar_caminho_absoluto(
            caminho_arquivo
        )

        # Instancia o objeto de print
        opcoes_print = PrintOptions()

        # caso os parâmetros não sejam None:
        if escala is not None:
            # define a escala
            opcoes_print.scale = escala
        if paginacao is not None:
            # define a paginação
            opcoes_print.page_ranges = paginacao
        if fundo is not None:
            # define o fundo
            opcoes_print.background = fundo
        if encolher_para_caber is not None:
            # define o ajuste de tamanho da página
            opcoes_print.shrink_to_fit = encolher_para_caber
        if orientacao is not None:
            # define a orientação da página
            orientacao_escolhida = opcoes_print.ORIENTATION_VALUES[orientacao]
            opcoes_print.orientation = orientacao_escolhida

        # coleta o hash base64 do print
        cache_base_64 = _navegador.print_page(opcoes_print)

        # inicia o gerenciador de contexto no arquivo de saída
        with open(caminho_arquivo_absoluto, 'wb') as arquivo_saida:
            # grava o hash base64 no arquivo de saida
            arquivo_saida.write(base64.b64decode(cache_base_64))

        # retorna True em caso de sucesso
        return True
    except:
        # retorna False em caso de falha
        return False


def fechar_janela(janela):
    """Fecha uma janela/aba do navegador automatizado."""
    _navegador.switch_to.window(janela)
    _navegador.close()


def fechar_janelas_menos_essa(id_janela):
    """Fecha todas as janelas/abas do
    navegador automatizado menos a informada."""
    lista_janelas = _navegador.window_handles
    for janela in lista_janelas:
        if janela != id_janela:
            _navegador.switch_to.window(janela)
            _navegador.close()


def encerrar_navegador():
    """Fecha a instância do navegador automatizado."""
    try:
        for janela in _navegador.window_handles:
            fechar_janela(janela)
        else:
            _navegador.quit()

        return True
    except:
        return False
